from ...http_client import HttpClient
from ...models import Webhook


class WebhookClient(HttpClient):
    """Client to interact with Webhooks

    **Preview**

    This is not yet finalised. It may change or be removed or be not yet
    accessible. This documentation is designed to give you an idea of features
    we'll be adding to our API in the future."""

    def __init__(self, **kwargs):
        self._url = "/air/webhooks"
        super().__init__(**kwargs)

    def create(self):
        """Initiate creation of a Webhook"""
        return WebhookCreate(self)

    def update(self, id_):
        """Initiate updating of a Webhook"""
        return WebhookUpdate(self, id_)

    def ping(self, id_):
        """Ping a webhook.

        Send a ping, a "fake event" notification, to a webhook.
        """
        url = f"{self._url}/{id_}/actions/ping"
        self.do_post(url)
        return None


class WebhookCreate(object):
    """Auxiliary class to provide methods for webhook creating related data"""

    def __init__(self, client):
        self._client = client

    def events(self, events):
        """Add events"""
        self._events = events
        return self

    def url(self, url):
        """Add URL"""
        self._url = url
        return self

    def execute(self):
        """POST /air/webhooks"""
        res = self._client.do_post(
            self._client._url,
            body={"data": {"events": self._events, "url": self._url}},
        )
        return Webhook.from_json(res["data"])


class WebhookUpdate(object):
    """Auxiliary class to provide methods for webhook updating related data"""

    def __init__(self, client, id):
        self._client = client
        self._id = id

    def active(self, active):
        """Set webhook to active"""
        self._active = active
        return self

    def execute(self):
        """PATCH /air/webhooks/{id}"""
        url = f"{self._client._url}/{self._id}"

        res = self._client.do_patch(
            url,
            body={"data": {"active": self._active}},
        )

        return Webhook.from_json(res["data"])



from typing import Any, Dict

from ...http_client import HttpClient, Pagination
from ...models import Order


class OrderClient(HttpClient):
    """Client to interact with Orders."""

    class InvalidSort(Exception):
        """Invalid sort option provided"""

    def __init__(self, **kwargs):
        """Instantiate an Order client."""
        self._url = "/air/orders"
        super().__init__(**kwargs)

    def get(self, id_):
        """GET /air/orders/:id."""
        res = self.do_get(f"{self._url}/{id_}")
        if res is not None:
            return Order.from_json(res["data"])

    def list(self, awaiting_payment=False, sort=None, limit=50):
        """GET /air/orders."""
        params: Dict[str, Any] = {"limit": limit}
        if sort:
            if sort not in ["pay_by", "-pay_by"]:
                raise OrderClient.InvalidSort(sort)
            params["sort"] = sort
        if awaiting_payment:
            params["awaiting_payment"] = "true"
        return Pagination(self, Order, params)

    def create(self):
        """Initiate creation of an Order."""
        return OrderCreate(self)

    def update(self, id_):
        """Initiate updating of an Order."""
        return OrderUpdate(self, id_)


class OrderCreate:
    """Auxiliary class to provide methods for order creation related data"""

    class InvalidSelectedOffersLength(Exception):
        """Invalid number of selected_offers provided"""

    class InvalidService(Exception):
        """Invalid service data provided"""

    class InvalidNumberOfPayments(Exception):
        """Invalid number of payments provided"""

    class InvalidPayment(Exception):
        """Invalid payment data provided"""

    class InvalidPaymentType(Exception):
        """Invalid payment type provided"""

    class InvalidNumberOfPassengers(Exception):
        """Invalid number of passengers provided"""

    class InvalidPassenger(Exception):
        """Invalid passenger data provided"""

    def __init__(self, client):
        self._client = client
        self._passengers = []
        self._payments = []
        self._selected_offers = []
        self._services = []
        self._payment_type = "instant"

    @staticmethod
    def _validate_payments(payments):
        """Validate number of payments and the data provided for each if any were given"""
        if len(payments) == 0:
            raise OrderCreate.InvalidNumberOfPayments(len(payments))
        for payment in payments:
            if set(payment.keys()) != set(["amount", "currency", "type"]):
                raise OrderCreate.InvalidPayment(payment)
            if payment["type"] not in ["arc_bsp_cash", "balance"]:
                raise OrderCreate.InvalidPaymentType(payment["type"])

    @staticmethod
    def _validate_services(services):
        """Validate the data provided for each service if any were given"""
        for service in services:
            if set(service.keys()) != set(["id", "quantity"]):
                raise OrderCreate.InvalidService(service)

    @staticmethod
    def _validate_selected_offers(selected_offers):
        """Validate number of selected_offers"""
        if len(selected_offers) != 1:
            raise OrderCreate.InvalidSelectedOffersLength(len(selected_offers))

    @staticmethod
    def _validate_passengers(passengers):
        """Validate passenger count and the data provided for each if any were given"""
        if len(passengers) == 0:
            raise OrderCreate.InvalidNumberOfPassengers(passengers)
        for passenger in passengers:
            if not (
                "born_on" in passenger
                or "email" in passenger
                or "family_name" in passenger
                or "gender" in passenger
                or "given_name" in passenger
                or "id" in passenger
                or "phone_number" in passenger
                or "title" in passenger
            ):
                raise OrderCreate.InvalidPassenger(passenger)

    def _instant_order(self):
        """Determine whether this order's type is instant"""
        return self._payment_type is not None and self._payment_type != "hold"

    def _build_order_payload(self):
        """Build order payload"""
        data = {
            "type": self._payment_type,
            "passengers": self._passengers,
            "services": self._services,
            "selected_offers": self._selected_offers,
        }

        if self._instant_order():
            data["payments"] = self._payments
        else:
            data["type"] = self._payment_type

        return data

    def hold(self):
        """Set payment type to 'hold'. If this isn't called the type is 'instant'"""
        self._payment_type = "hold"
        return self

    def selected_offers(self, selected_offers):
        """Set selected_offers"""
        OrderCreate._validate_selected_offers(selected_offers)
        self._selected_offers = selected_offers
        return self

    def services(self, services):
        """Set services"""
        # TODO(nlopes): this should be its own type to ensure the user *only* passes valid
        # data
        OrderCreate._validate_services(services)
        self._services = services
        return self

    def payments(self, payments):
        """Set payment method"""
        # TODO(nlopes): this should be its own type to ensure the user *only* passes valid
        # data
        OrderCreate._validate_payments(payments)
        self._payments = payments
        return self

    def passengers(self, passengers):
        """Set passenger information for all passengers that will be travelling"""
        # TODO(nlopes): this should be its own type to ensure the user *only* passes valid
        # data
        OrderCreate._validate_passengers(passengers)
        self._passengers = passengers
        return self

    def execute(self):
        """POST /air/orders - trigger the call to create the order"""
        OrderCreate._validate_passengers(self._passengers)
        if self._instant_order():
            OrderCreate._validate_payments(self._payments)
        OrderCreate._validate_services(self._services)
        OrderCreate._validate_selected_offers(self._selected_offers)
        res = self._client.do_post(
            self._client._url,
            body={"data": self._build_order_payload()},
        )
        return Order.from_json(res["data"])


class OrderUpdate:
    """Auxiliary class to provide methods for order update related data"""

    class InvalidMetadata(Exception):
        """Invalid metadata data provided"""

    def __init__(self, client, id_):
        self._client = client
        self._id = id_
        self._metadata = {}

    @staticmethod
    def _validate_metadata(metadata):
        """Validate structure of Metadata"""
        if type(metadata) is not dict:
            raise OrderUpdate.InvalidMetadata(metadata)

    def metadata(self, metadata):
        """Set the Order Metadata."""
        self._metadata = metadata
        OrderUpdate._validate_metadata(self._metadata)
        return self

    def execute(self):
        """PATCH /air/orders/{:id} - trigger the call to update the order."""
        OrderUpdate._validate_metadata(self._metadata)

        url = f"{self._client._url}/{self._id}"

        res = self._client.do_patch(
            url,
            body={"data": {"metadata": self._metadata}},
        )
        if res is not None:
            return Order.from_json(res["data"])



import pytest

from duffel_api.api import PaymentClient

from .fixtures import fixture


def test_create_payment(requests_mock):
    url = "air/payments"
    with fixture("create-payment", url, requests_mock.post, 201) as client:
        payment_details = {
            "type": "balance",
            "currency": "GBP",
            "amount": "30.20",
        }
        payment = (
            client.payments.create()
            .order("order-id")
            .payment(payment_details)
            .execute()
        )
        assert payment.id == "pay_00009hthhsUZ8W4LxQgkjo"
        assert payment.type == "balance"
        assert payment.amount == "30.20"
        assert payment.currency == "GBP"


def test_create_payment_with_invalid_payment_details(requests_mock):
    url = "air/payments"
    with fixture("create-payment", url, requests_mock.post, 422) as client:
        payment_details = {
            "currency": "GBP",
            "amount": "30.20",
        }
        payments_create_client = client.payments.create().order("order-id")
        with pytest.raises(
            PaymentClient.InvalidPayment,
            match="{'currency': 'GBP', 'amount': '30.20'}",
        ):
            payments_create_client.payment(payment_details).execute()
        payment_details["type"] = "credit"
        with pytest.raises(PaymentClient.InvalidPaymentType, match="credit"):
            payments_create_client.payment(payment_details).execute()



from ...http_client import HttpClient
from ...models import PaymentIntent


class PaymentIntentClient(HttpClient):
    """To begin the process of collecting a card payment from your customer, you
    need to create a Payment Intent.

    The Payment Intent will contain a client_token that you use to collect the
    card payment in your application.

    If the Payment Intent is created in test mode you should use a test card.
    """

    def __init__(self, **kwargs):
        self._url = "/payments/payment_intents"
        super().__init__(**kwargs)

    def create(self):
        """Initiate creation of a Payment Intent"""
        return PaymentIntentCreate(self)

    def get(self, id_):
        """Get a single Payment Intent.

        You should use this API to get the complete, up-to-date information
        about a Payment Intent.
        """
        res = self.do_get(f"{self._url}/{id_}")
        if res is not None:
            return PaymentIntent.from_json(res["data"])

    def confirm(self, id_):
        """Confirm a Payment Intent

        Once you've successfully collected the customer's card details, using
        the client_token from when you first created the Payment Intent, you
        then need to confirm it using this endpoint.

        Once confirmed, the amount charged to your customer's card will be added
        to your Balance (minus any Duffel Payment fees).
        """
        res = self.do_post(f"{self._url}/{id_}/actions/confirm")
        if res is not None:
            return PaymentIntent.from_json(res["data"])


class PaymentIntentCreate:
    """Auxiliary class to provide methods for creating payment intents"""

    class InvalidPayment(Exception):
        """Invalid payment data provided"""

    def __init__(self, client):
        self._client = client
        self._amount = None
        self._currency = None

    @staticmethod
    def _validate_payment_keys(payment_details):
        """Validate that payment keys are correct"""
        if set(payment_details.keys()) != set(["amount", "currency"]):
            raise PaymentIntentCreate.InvalidPayment(payment_details)

    def payment(self, payment_details):
        """Set amount and currency"""
        PaymentIntentCreate._validate_payment_keys(payment_details)

        self._amount = payment_details["amount"]
        self._currency = payment_details["currency"]

        return self

    def execute(self):
        """POST /payments/payment_intents"""
        if self._amount is None:
            raise PaymentIntentCreate.InvalidPayment()

        if self._currency is None:
            raise PaymentIntentCreate.InvalidPayment()

        res = self._client.do_post(
            self._client._url,
            body={
                "data": {
                    "amount": self._amount,
                    "currency": self._currency,
                }
            },
        )
        return PaymentIntent.from_json(res["data"])



"""Auxiliary contexts for handling fixtures"""
import json
from contextlib import contextmanager

from duffel_api import Duffel


@contextmanager
def fixture(name, url_path, mock, status_code, with_response=True):
    """Yields a Duffel client instnace.

    It responds to a given `url_path` under `mock`."""
    url = "http://someaddress"
    with open(f"tests/fixtures/{name}.json") as fh:
        if with_response:
            response = json.loads(fh.read())
            uri = f"{url}/{url_path}"
            mock(uri, complete_qs=True, json=response, status_code=status_code)
            yield Duffel(access_token="some_token", api_url=url)
        else:
            uri = f"{url}/{url_path}"
            mock(uri, complete_qs=True, status_code=status_code)
            yield Duffel(access_token="some_token", api_url=url)


@contextmanager
def raw_fixture(name):
    """Yields an instance of the fixture that has been loaded as JSON."""
    with open(f"tests/fixtures/{name}.json") as fh:
        yield json.load(fh)



from duffel_api.models import Order

from .fixtures import raw_fixture


def test_order_model_parsing():
    name = "get-order-by-id"
    with raw_fixture(name) as fixture:
        assert Order.from_json(fixture["data"])



import datetime

from .fixtures import fixture


def test_get_order_cancellation_by_id(requests_mock):
    url = "air/order_cancellations/id"
    with fixture("get-order-cancellation-by-id", url, requests_mock.get, 200) as client:
        order_cancellation = client.order_cancellations.get("id")
        assert order_cancellation.id == "ore_00009qzZWzjDipIkqpaUAj"
        assert order_cancellation.order_id == "ord_00009hthhsUZ8W4LxQgkjo"
        assert order_cancellation.refund_amount == "90.80"
        assert order_cancellation.refund_currency == "GBP"
        assert order_cancellation.refund_to == "arc_bsp_cash"


def test_get_order_cancellations(requests_mock):
    # We need a way to ensure pagination finished in a mocking environment
    end_pagination_url = (
        "http://someaddress/air/order_cancellations?limit=50"
        + "&after=g2wAAAACbQAAABBBZXJvbWlzdC1LaGFya2l2bQAAAB%3D&order_id=order-id"
    )
    end_pagination_response = {"meta": {"after": None}, "data": []}
    requests_mock.get(
        end_pagination_url, complete_qs=True, json=end_pagination_response
    )

    url = "air/order_cancellations?limit=50&order_id=order-id"
    with fixture("get-order-cancellations", url, requests_mock.get, 200) as client:
        paginated_order_cancellations = client.order_cancellations.list("order-id")
        order_cancellations = list(paginated_order_cancellations)
        assert len(order_cancellations) == 1
        order_cancellation = order_cancellations[0]
        assert order_cancellation.id == "ore_00009qzZWzjDipIkqpaUAj"


def test_create_order_cancellation(requests_mock):
    url = "air/order_cancellations"
    with fixture("create-order-cancellation", url, requests_mock.post, 201) as client:
        cancellation = client.order_cancellations.create("order-id")
        assert cancellation.id == "ore_00009qzZWzjDipIkqpaUAj"
        assert cancellation.refund_to == "arc_bsp_cash"
        assert cancellation.refund_currency == "GBP"
        assert cancellation.refund_amount == "90.80"


def test_confirm_order_cancellation(requests_mock):
    url = "air/order_cancellations/some-id/actions/confirm"
    with fixture("confirm-order-cancellation", url, requests_mock.post, 200) as client:
        cancellation = client.order_cancellations.confirm("some-id")
        assert cancellation.id == "ore_00009qzZWzjDipIkqpaUAj"
        assert cancellation.refund_to == "arc_bsp_cash"
        assert cancellation.refund_currency == "GBP"
        assert cancellation.refund_amount == "90.80"
        assert cancellation.confirmed_at == datetime.datetime(
            2020, 1, 17, 11, 51, 43, 114803
        )



from datetime import datetime, date

from .fixtures import fixture


def test_create_order_change_request(requests_mock):
    url = "air/order_change_requests"
    with fixture("create-order-change-request", url, requests_mock.post, 201) as client:
        slices = {
            "add": [
                {
                    "cabin_class": "economy",
                    "departure_date": "2020-04-24",
                    "destination": "JFK",
                    "origin": "LHR",
                }
            ],
            "remove": [{"slice_id": "sli_00009htYpSCXrwaB9Dn123"}],
        }
        creation = client.order_change_requests.create("order-id").slices(slices)
        change_request = creation.execute()
        assert change_request.created_at == datetime(2020, 1, 17, 10, 12, 14, 545000)

        assert change_request.id == "ocr_0000A3bQP9RLVfNUcdpLpw"
        assert change_request.live_mode is True
        assert change_request.order_id == "ord_0000A3bQ8FJIQoEfuC07n6"
        assert isinstance(change_request.slices.add, type([]))
        assert change_request.slices.add[0].cabin_class == "economy"
        assert change_request.slices.add[0].departure_date == date(2020, 4, 24)
        assert change_request.slices.add[0].destination == "JFK"
        assert change_request.slices.add[0].origin == "LHR"
        assert isinstance(change_request.slices.remove, type([]))
        assert change_request.slices.remove[0].slice_id == "sli_00009htYpSCXrwaB9Dn123"
        assert change_request.updated_at == datetime(2020, 1, 17, 10, 12, 14, 545000)


def test_get_order_change_request(requests_mock):
    url = "air/order_change_requests/order-id"
    with fixture(
        "get-order-change-request-by-id", url, requests_mock.get, 200
    ) as client:
        change_request = client.order_change_requests.get("order-id")
        assert change_request.created_at == datetime(2020, 1, 17, 10, 12, 11, 634000)
        assert change_request.id == "ocr_0000A3bQP9RLVfNUcdpLpw"
        assert change_request.live_mode is True
        assert change_request.order_id == "ord_0000A3bQ8FJIQoEfuC07n6"
        assert isinstance(change_request.order_change_offers, type([]))
        assert (
            change_request.order_change_offers[0].slices.add[0].id
            == "sli_00009htYpSCXrwaB9Dn123"
        )
        assert isinstance(change_request.order_change_offers[0].slices.remove, type([]))
        assert (
            change_request.order_change_offers[0].slices.remove[0].id
            == "sli_00009htYpSCXrwaB9Dn123"
        )
        assert change_request.order_change_offers[0].updated_at == datetime(
            2020, 1, 17, 10, 12, 14, 545000
        )
        assert isinstance(change_request.slices.add, type([]))
        assert change_request.slices.add[0].cabin_class == "economy"
        assert change_request.slices.add[0].departure_date == date(2020, 4, 24)
        assert change_request.slices.add[0].destination == "JFK"
        assert change_request.slices.add[0].origin == "LHR"
        assert isinstance(change_request.slices.remove, type([]))
        assert change_request.slices.remove[0].slice_id == "sli_00009htYpSCXrwaB9Dn123"
        assert change_request.updated_at == datetime(2020, 1, 17, 10, 12, 11, 634000)



from duffel_api.models import Airline

from .fixtures import raw_fixture


def test_airline_model_parsing():
    name = "get-airline-by-id"
    with raw_fixture(name) as fixture:
        assert Airline.from_json(fixture["data"])



from .fixtures import fixture


def test_get_seat_maps(requests_mock):
    url = "air/seat_maps?offer_id=offer-id"
    with fixture("get-seat-maps", url, requests_mock.get, 200) as client:
        seat_maps = client.seat_maps.get("offer-id")
        assert len(seat_maps) == 1
        seat_map = seat_maps[0]
        assert len(seat_map.cabins) == 1
        assert seat_map.cabins[0].aisles == 2
        assert len(seat_map.cabins[0].rows) == 7
        element = seat_map.cabins[0].rows[0].sections[0].elements[0]
        assert element.designator == "1A"
        assert element.type == "seat"
        assert element.available_services[0].id == "ase_00009UhD4ongolulWAAA1A"
        assert element.available_services[0].total_amount == "30.00"



import pytest

from datetime import datetime

from duffel_api.api import PaymentIntentCreate

from .fixtures import fixture


def test_create_payment_intent(requests_mock):
    url = "payments/payment_intents"
    with fixture("create-payment-intent", url, requests_mock.post, 201) as client:
        payment_intent_details = {
            "amount": "30.20",
            "currency": "GBP",
        }
        creation = client.payment_intents.create()
        payment_intent = creation.payment(payment_intent_details).execute()

        assert payment_intent.amount == "30.20"
        assert (
            payment_intent.client_token
            == "eyJjbGllbnRfc2VjcmV0IjoicGlfM0s2dFMyQW5rMVRkeXJvRDA0eVVzM1BxX3NlY3JldF9VY"
            + "zV3VjRYU0hUc2VHWHlSOHhpTTNsdFdjIiwicHVibGlzaGFibGVfa2V5IjoicGtfdGVzdF81MUl"
            + "0Q3YwQW5rMVRkeXJvRGxFTGExeGdZZkpBMUdUdzh2VmRRVmVRUEl3S1hiVjk5MVNnNnJpY1Nac"
            + "1hJWGtTZ2VGWGF1c1RBcVVQSkhUNXNMZWpIdFlIUjAwVEVESVh4cnMifQ=="
        )
        assert payment_intent.confirmed_at is None
        assert payment_intent.created_at == datetime(2020, 4, 11, 15, 48, 11, 642000)
        assert payment_intent.currency == "GBP"
        assert payment_intent.fees_amount is None
        assert payment_intent.fees_currency is None
        assert payment_intent.id == "pit_0000AEQIx7Swvd9nmMNGkq"
        assert payment_intent.live_mode is True
        assert payment_intent.net_amount is None
        assert payment_intent.net_currency is None
        assert payment_intent.refunds == []
        assert payment_intent.status == "requires_payment_method"
        assert payment_intent.updated_at == datetime(2020, 4, 11, 15, 48, 11, 642000)


def test_create_payment_intent_with_invalid_data(requests_mock):
    url = "payments/payment_intents"
    with fixture("create-payment-intent", url, requests_mock.post, 422) as client:
        creation = client.payment_intents.create()
        with pytest.raises(PaymentIntentCreate.InvalidPayment):
            creation.payment({})

        with pytest.raises(PaymentIntentCreate.InvalidPayment):
            creation.payment({"amount": "100.00"})

        with pytest.raises(PaymentIntentCreate.InvalidPayment):
            creation.payment({"currency": "AUD"})


def test_get_payment_intent(requests_mock):
    url = "payments/payment_intents/id"
    with fixture("get-payment-intent-by-id", url, requests_mock.get, 200) as client:
        payment_intent = client.payment_intents.get("id")

        assert payment_intent.amount == "30.20"
        assert (
            payment_intent.client_token
            == "eyJjbGllbnRfc2VjcmV0IjoicGlfM0s2dFMyQW5rMVRkeXJvRDA0eVVzM1BxX3NlY3JldF9VY"
            + "zV3VjRYU0hUc2VHWHlSOHhpTTNsdFdjIiwicHVibGlzaGFibGVfa2V5IjoicGtfdGVzdF81MUl"
            + "0Q3YwQW5rMVRkeXJvRGxFTGExeGdZZkpBMUdUdzh2VmRRVmVRUEl3S1hiVjk5MVNnNnJpY1Nac"
            + "1hJWGtTZ2VGWGF1c1RBcVVQSkhUNXNMZWpIdFlIUjAwVEVESVh4cnMifQ=="
        )
        assert payment_intent.confirmed_at is None
        assert payment_intent.card_country_code == "GB"
        assert payment_intent.card_last_four_digits == "0000"
        assert payment_intent.card_network == "visa"
        assert payment_intent.created_at == datetime(2020, 4, 11, 15, 48, 11, 642000)
        assert payment_intent.currency == "GBP"
        assert payment_intent.fees_amount is None
        assert payment_intent.fees_currency is None
        assert payment_intent.id == "pit_0000AEQIx7Swvd9nmMNGkq"
        assert payment_intent.live_mode is True
        assert payment_intent.net_amount is None
        assert payment_intent.net_currency is None
        assert payment_intent.refunds == []
        assert payment_intent.status == "succeeded"
        assert payment_intent.updated_at == datetime(2020, 4, 11, 15, 48, 39, 246000)


def test_confirm_payment_intent(requests_mock):
    url = "payments/payment_intents/id/actions/confirm"
    with fixture("confirm-payment-intent", url, requests_mock.post, 200) as client:
        payment_intent = client.payment_intents.confirm("id")

        assert payment_intent.amount == "30.20"
        assert (
            payment_intent.client_token
            == "eyJjbGllbnRfc2VjcmV0IjoicGlfM0s2dFMyQW5rMVRkeXJvRDA0eVVzM1BxX3NlY3JldF9VY"
            + "zV3VjRYU0hUc2VHWHlSOHhpTTNsdFdjIiwicHVibGlzaGFibGVfa2V5IjoicGtfdGVzdF81MUl"
            + "0Q3YwQW5rMVRkeXJvRGxFTGExeGdZZkpBMUdUdzh2VmRRVmVRUEl3S1hiVjk5MVNnNnJpY1Nac"
            + "1hJWGtTZ2VGWGF1c1RBcVVQSkhUNXNMZWpIdFlIUjAwVEVESVh4cnMifQ=="
        )
        assert payment_intent.confirmed_at == datetime(2020, 4, 11, 15, 48, 30, 321000)
        assert payment_intent.card_country_code == "GB"
        assert payment_intent.card_last_four_digits == "0000"
        assert payment_intent.card_network == "visa"
        assert payment_intent.created_at == datetime(2020, 4, 11, 15, 48, 11, 642000)
        assert payment_intent.currency == "GBP"
        assert payment_intent.fees_amount == "0.42"
        assert payment_intent.fees_currency == "GBP"
        assert payment_intent.id == "pit_0000AEQIx7Swvd9nmMNGkq"
        assert payment_intent.live_mode is True
        assert payment_intent.net_amount == "29.78"
        assert payment_intent.net_currency == "GBP"
        assert payment_intent.refunds == []
        assert payment_intent.status == "succeeded"
        assert payment_intent.updated_at == datetime(2020, 4, 11, 15, 48, 39, 246000)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from duffel_api.utils import get_and_transform


@dataclass
class OrderCancellation:
    """To cancel an order, you'll need to create an order cancellation,
    check the refund_amount returned, and, if you're happy to go ahead and
    cancel the order.

    The refund specified by refund_amount, if any, will be returned to your
    original payment method (i.e. your Duffel balance).
    You'll then need to refund your customer (e.g. back to their credit/debit card).
    """

    id: str
    order_id: str
    live_mode: bool
    expires_at: datetime
    refund_amount: str
    refund_currency: str
    refund_to: str
    confirmed_at: Optional[datetime]
    created_at: datetime

    allowed_refund_types = [
        "arc_bsp_cash",
        "balance",
        "card",
        "voucher",
        "awaiting_payment",
    ]

    class InvalidRefundType(Exception):
        """Invalid refund type provided"""

    def __post_init__(self):
        if self.refund_to not in OrderCancellation.allowed_refund_types:
            raise OrderCancellation.InvalidRefundType(self.refund_to)

        return self

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            order_id=json["order_id"],
            live_mode=json["live_mode"],
            expires_at=parse_datetime(json["expires_at"]),
            refund_amount=json["refund_amount"],
            refund_currency=json["refund_currency"],
            refund_to=json["refund_to"],
            confirmed_at=get_and_transform(json, "confirmed_at", parse_datetime),
            created_at=datetime.strptime(json["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
        )


def parse_datetime(value: str) -> datetime:
    # There are inconsistent formats used for this field depending on the
    # endpoint
    if len(value) == 20:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ")
    else:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")



from duffel_api.models import OfferRequest

from .fixtures import raw_fixture


def test_offer_request_model_parsing():
    name = "get-offer-request-by-id"
    with raw_fixture(name) as fixture:
        assert OfferRequest.from_json(fixture["data"])



from duffel_api.models import OrderChangeOffer

from .fixtures import raw_fixture


def test_order_change_offer_model_parsing():
    name = "get-order-change-offer-by-id"
    with raw_fixture(name) as fixture:
        assert OrderChangeOffer.from_json(fixture["data"])



from duffel_api.models import Aircraft

from .fixtures import raw_fixture


def test_aircraft_model_parsing():
    name = "get-aircraft-by-id"
    with raw_fixture(name) as fixture:
        assert Aircraft.from_json(fixture["data"])



import pytest

from duffel_api.http_client import HttpClient, ApiError


def test_http_client(requests_mock):
    requests_mock.get("http://someaddress/api/stuff", json={})
    client = HttpClient("some_token", "http://someaddress", "v1")
    assert client.do_get("/api/stuff") == {}


def test_http_client_error(requests_mock):
    error = {
        "meta": {"status": 500, "request_id": "FmXeZifDA60QOlgAAODB"},
        "errors": [
            {
                "type": "airline_error",
                "title": "Unexpected Airline Error",
                "source": "iberia",
                "message": "The airline responded with an unexpected error.",
                "documentation_url": "https://duffel.com/docs/api/overview/errors",
                "code": "airline_unknown",
            }
        ],
    }
    requests_mock.get("http://someaddress/api/stuff", json=error, status_code=500)
    client = HttpClient("some_token", "http://someaddress", "v1")
    with pytest.raises(
        ApiError, match="The airline responded with an unexpected error"
    ):
        client.do_get("/api/stuff")



"""Entry point to the library"""
from .api import (
    AircraftClient,
    AirportClient,
    AirlineClient,
    OfferRequestClient,
    OfferClient,
    OrderCancellationClient,
    OrderChangeRequestClient,
    OrderClient,
    OrderChangeClient,
    OrderChangeOfferClient,
    PaymentClient,
    PaymentIntentClient,
    SeatMapClient,
    WebhookClient,
)


def lazy_property(func):
    """Decorator that makes a property lazy-evaluated"""
    attr_name = "_lazy_" + func.__name__

    @property
    def _lazy_property(self):
        """Function for property lazy-evaluate decorator"""
        if not hasattr(self, attr_name):
            # The attr doesn't exist yet, so call the function (`func`) that has
            # the decorator. The result of that call is set in the attr.
            #
            # It's okay to set (read: memoise) the client as an attr since we
            # each call to the client's functions (e.g. `list`) are not
            # memoised.
            setattr(self, attr_name, func(self))

        # Get the result of the function call that was set in the attr.
        return getattr(self, attr_name)

    return _lazy_property


class Duffel:
    """Client to the entire API"""

    def __init__(self, **kwargs):
        # TODO(nlopes): I really don't like how I've built this- we shouldn't keep
        # instantiating the HttpClient through class inheritance.
        # We should (maybe!) have a singleton pattern on HttpClient and use
        # composition in all of these instead.

        # Keep this as we use it when doing the lazy-evaluation of the different
        # clients
        self._kwargs = kwargs

    @lazy_property
    def aircraft(self):
        """Aircraft API - /air/aircraft"""
        return AircraftClient(**self._kwargs)

    @lazy_property
    def airports(self):
        """Airports API - /air/airports"""
        return AirportClient(**self._kwargs)

    @lazy_property
    def airlines(self):
        """Airlines API - /air/airlines"""
        return AirlineClient(**self._kwargs)

    @lazy_property
    def offer_requests(self):
        """Offer Requests API - /air/offer_requests"""
        return OfferRequestClient(**self._kwargs)

    @lazy_property
    def offers(self):
        """Offers API - /air/offers"""
        return OfferClient(**self._kwargs)

    @lazy_property
    def orders(self):
        """Orders API - /air/orders"""
        return OrderClient(**self._kwargs)

    @lazy_property
    def order_cancellations(self):
        """Order Cancellations API - /air/order_cancellations"""
        return OrderCancellationClient(**self._kwargs)

    @lazy_property
    def order_changes(self):
        """Order Changes API - /air/order_changes"""
        return OrderChangeClient(**self._kwargs)

    @lazy_property
    def order_change_offers(self):
        """Order Change Offers API - /air/order_change_offers"""
        return OrderChangeOfferClient(**self._kwargs)

    @lazy_property
    def order_change_requests(self):
        """Order Change Requests API - /air/order_change_requests"""
        return OrderChangeRequestClient(**self._kwargs)

    @lazy_property
    def payment_intents(self):
        """Payment Intents API - /payments/payment_intents"""
        return PaymentIntentClient(**self._kwargs)

    @lazy_property
    def payments(self):
        """Payments API - /air/payments"""
        return PaymentClient(**self._kwargs)

    @lazy_property
    def seat_maps(self):
        """Seat Maps API - /air/seat_maps"""
        return SeatMapClient(**self._kwargs)

    @lazy_property
    def webhooks(self):
        """Webhooks API - /air/webhooks (Preview)"""
        return WebhookClient(**self._kwargs)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Sequence

from duffel_api.models import Refund
from duffel_api.utils import get_and_transform


@dataclass
class PaymentIntent:
    """To begin the process of collecting a card payment from your customer, you
    need to create a Payment Intent.

    The Payment Intent will contain a client_token that you use to collect the
    card payment in your application.

    If the Payment Intent is created in test mode you should use a test card.
    """

    id: str
    live_mode: bool
    amount: str
    currency: str
    net_amount: Optional[str]
    net_currency: Optional[str]
    fees_amount: Optional[str]
    fees_currency: Optional[str]
    client_token: str
    card_network: Optional[str]
    card_last_four_digits: Optional[str]
    card_country_code: Optional[str]
    status: str
    refunds: Sequence[Refund]
    confirmed_at: Optional[datetime]
    created_at: datetime
    updated_at: datetime

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            live_mode=json["live_mode"],
            amount=json["amount"],
            currency=json["currency"],
            net_amount=json.get("net_amount"),
            net_currency=json.get("net_currency"),
            fees_amount=json.get("fees_amount"),
            fees_currency=json.get("fees_currency"),
            client_token=json["client_token"],
            card_network=json.get("card_network"),
            card_last_four_digits=json.get("card_last_four_digits"),
            card_country_code=json.get("card_country_code"),
            status=json["status"],
            refunds=get_and_transform(
                json,
                "refunds",
                lambda value: [Refund.from_json(refund) for refund in value],
                [],
            ),
            confirmed_at=get_and_transform(json, "confirmed_at", parse_datetime),
            created_at=datetime.strptime(json["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            updated_at=datetime.strptime(json["updated_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
        )


def parse_datetime(value: str) -> datetime:
    # There are inconsistent formats used for this field depending on the
    # endpoint
    if len(value) == 20:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ")
    else:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Sequence

from duffel_api.utils import get_and_transform


@dataclass
class City:
    """The metropolitan area where the airport is located.
    Only present for airports which are registered with IATA as
    belonging to a metropolitan area.
    """

    id: str
    name: str
    iata_code: str
    iata_country_code: str

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            name=json["name"],
            iata_code=json["iata_code"],
            iata_country_code=json["iata_country_code"],
        )


@dataclass
class Airport:
    """Airports are used to identify origins and destinations in journey
    slices"""

    id: str
    name: str
    iata_code: Optional[str]
    icao_code: Optional[str]
    iata_country_code: str
    latitude: float
    longitude: float
    time_zone: str
    city: Optional[City]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            name=json["name"],
            iata_code=json.get("iata_code"),
            icao_code=json.get("icao_code"),
            iata_country_code=json["iata_country_code"],
            latitude=json["latitude"],
            longitude=json["longitude"],
            time_zone=json["time_zone"],
            city=get_and_transform(json, "city", City.from_json),
        )


@dataclass
class Place:
    """The city or airport"""

    id: str
    name: str
    type: str
    iata_city_code: Optional[str]
    iata_country_code: str
    latitude: Optional[float]
    longitude: Optional[float]
    icao_code: Optional[str]
    time_zone: Optional[str]
    city_name: Optional[str]
    city: Optional[City]
    airports: Optional[Sequence[Airport]]

    allowed_types = ["airport", "city"]

    class InvalidType(Exception):
        """Invalid type of place"""

    def __post_init__(self):
        if self.type not in Place.allowed_types:
            raise Place.InvalidType(self.type)

        return self

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            name=json["name"],
            type=json["type"],
            iata_city_code=json.get("iata_city_code"),
            iata_country_code=json["iata_country_code"],
            latitude=json.get("latitude"),
            longitude=json.get("longitude"),
            icao_code=json.get("icao_code"),
            time_zone=json.get("time_zone"),
            city_name=json.get("city_name"),
            city=get_and_transform(json, "city", City.from_json),
            airports=get_and_transform(
                json,
                "airports",
                lambda value: [Airport.from_json(airport) for airport in value],
            ),
        )


@dataclass
class Refund:
    """A Refund allows you to refund money that you had collected from a customer with a
    Payment Intent. You're able to do partial refunds and also able to do multiple
    refunds for the same Payment Intent.
    """

    id: str
    live_mode: bool
    payment_intent_id: str
    amount: str
    currency: str
    net_amount: Optional[str]
    net_currency: Optional[str]
    status: str
    destination: str
    arrival: str
    created_at: datetime
    updated_at: datetime

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            live_mode=json["live_mode"],
            payment_intent_id=json["payment_intent_id"],
            amount=json["amount"],
            currency=json["currency"],
            net_amount=json.get("net_amount"),
            net_currency=json.get("net_currency"),
            status=json["status"],
            destination=json["destination"],
            arrival=json["arrival"],
            created_at=datetime.strptime(json["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            updated_at=datetime.strptime(json["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
        )



import pytest

from duffel_api.api import OfferRequestCreate

from .fixtures import fixture


def test_get_offer_request_by_id(requests_mock):
    url = "air/offer_requests/id"
    with fixture("get-offer-request-by-id", url, requests_mock.get, 200) as client:
        offer_request = client.offer_requests.get("id")
        assert offer_request.id == "orq_00009hjdomFOCJyxHG7k7k"
        assert len(offer_request.slices) == 1
        assert len(offer_request.passengers) == 1
        slice = offer_request.slices[0]
        assert slice.origin_type == "airport"


def test_get_offer_requests(requests_mock):
    # We need a way to ensure pagination finished in a mocking environment
    end_pagination_url = (
        "http://someaddress/air/offer_requests?limit=50"
        + "&after=g2wAAAACbQAAABBBZXJvbWlzdC1LaGFya2l2bQAAAB%3D"
    )
    end_pagination_response = {"meta": {"after": None}, "data": []}
    requests_mock.get(
        end_pagination_url, complete_qs=True, json=end_pagination_response
    )

    url = "air/offer_requests?limit=50"
    with fixture("get-offer-requests", url, requests_mock.get, 200) as client:
        paginated_offer_requests = client.offer_requests.list()
        offer_requests = list(paginated_offer_requests)
        assert len(offer_requests) == 1
        offer_request = offer_requests[0]
        assert offer_request.id == "orq_00009hjdomFOCJyxHG7k7k"


def test_create_offer_request(requests_mock):
    url = "air/offer_requests?return_offers=false"
    with fixture("create-offer-request", url, requests_mock.post, 201) as client:
        passengers = [{"type": "adult"}]
        slices = [
            {
                "departure_date": "2100-02-27",
                "destination": "LGW",
                "origin": "LIS",
            }
        ]
        offer_request = (
            client.offer_requests.create()
            .passengers(passengers)
            .slices(slices)
            .execute()
        )
        assert offer_request.id == "orq_00009hjdomFOCJyxHG7k7k"
        assert len(offer_request.offers) == 1
        offer = offer_request.offers[0]
        assert offer.id == "off_00009htYpSCXrwaB9DnUm0"


def test_create_offer_request_with_invalid_data(requests_mock):
    url = "air/offer_requests?return_offers=false"
    with fixture("create-offer-request", url, requests_mock.post, 422) as client:
        creation = client.offer_requests.create()
        with pytest.raises(OfferRequestCreate.InvalidNumberOfPassengers):
            creation.execute()

        with pytest.raises(OfferRequestCreate.InvalidPassenger):
            creation.passengers([{}]).execute()

        with pytest.raises(OfferRequestCreate.InvalidCabinClass):
            creation.cabin_class("invalid").execute()

        passengers = [{"type": "adult"}]

        creation = creation.passengers(passengers)
        with pytest.raises(OfferRequestCreate.InvalidNumberOfSlices):
            creation.execute()
        with pytest.raises(OfferRequestCreate.InvalidSlice):
            creation.slices([{}]).execute()



from duffel_api.models import Payment


def test_payment_model_parsing():
    json = {
        "type": "balance",
        "live_mode": False,
        "id": "pay_00009hthhsUZ8W4LxQgkjo",
        "currency": "GBP",
        "created_at": "2020-04-11T15:48:11.642Z",
        "amount": "30.20",
    }
    assert Payment.from_json(json)



from duffel_api.models import Airport

from .fixtures import raw_fixture


def test_airport_model_parsing():
    name = "get-airport-by-id"
    with raw_fixture(name) as fixture:
        assert Airport.from_json(fixture["data"])



from duffel_api.models import Refund

from .fixtures import raw_fixture


def test_refund_model_parsing():
    name = "get-refund-by-id"
    with raw_fixture(name) as fixture:
        assert Refund.from_json(fixture["data"])



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Sequence


@dataclass
class Webhook:
    """
    Webhooks are used to automatically receive notifications of events that
    happen. For example, when an order has a schedule change.
    """

    id: str
    live_mode: bool
    active: bool
    created_at: datetime
    events: Sequence[str]
    updated_at: datetime
    url: str
    secret: Optional[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            live_mode=json["live_mode"],
            active=json["active"],
            created_at=datetime.strptime(json["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            events=json["events"],
            updated_at=datetime.strptime(json["updated_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            url=json["url"],
            secret=json.get("secret"),
        )



from .supporting.aircraft import AircraftClient
from .supporting.airports import AirportClient
from .supporting.airlines import AirlineClient
from .booking.offer_requests import OfferRequestClient, OfferRequestCreate
from .booking.offers import OfferClient
from .booking.orders import OrderClient, OrderCreate, OrderUpdate
from .booking.order_cancellations import OrderCancellationClient
from .booking.order_changes import OrderChangeClient
from .booking.order_change_offers import OrderChangeOfferClient
from .booking.order_change_requests import OrderChangeRequestClient
from .booking.payments import PaymentClient
from .booking.seat_maps import SeatMapClient
from .duffel_payments.payment_intents import PaymentIntentClient, PaymentIntentCreate
from .notifications.webhooks import WebhookClient

__all__ = [
    "AircraftClient",
    "AirportClient",
    "AirlineClient",
    "OfferRequestClient",
    "OfferRequestCreate",
    "OfferClient",
    "OrderChangeClient",
    "OrderChangeOfferClient",
    "OrderChangeRequestClient",
    "OrderClient",
    "OrderCreate",
    "OrderUpdate",
    "OrderCancellationClient",
    "PaymentClient",
    "PaymentIntentClient",
    "PaymentIntentCreate",
    "SeatMapClient",
    "WebhookClient",
]



from duffel_api.models import OrderChange

from .fixtures import raw_fixture


def test_order_change_model_parsing():
    name = "get-order-change-by-id"
    with raw_fixture(name) as fixture:
        assert OrderChange.from_json(fixture["data"])



from duffel_api.models import LoyaltyProgrammeAccount


def test_loyalty_programme_account_model_parsing():
    json = {"airline_iata_code": "BA", "account_number": "12901014"}
    assert LoyaltyProgrammeAccount.from_json(json)



import datetime

import pytest

from duffel_api.api import OrderChangeClient

from .fixtures import fixture


def test_get_order_change_by_id(requests_mock):
    url = "air/order_changes/ocr_0000A3tQSmKyqOrcySrGbo"
    with fixture("get-order-change-by-id", url, requests_mock.get, 200) as client:
        order_change = client.order_changes.get("ocr_0000A3tQSmKyqOrcySrGbo")

        assert order_change.id == "ocr_0000A3tQSmKyqOrcySrGbo"
        assert order_change.change_total_amount == "30.50"
        assert order_change.change_total_currency == "GBP"
        assert order_change.confirmed_at == datetime.datetime(2020, 1, 17, 11, 51, 43)
        assert order_change.created_at == datetime.datetime(
            2020, 4, 11, 15, 48, 11, 642000
        )
        assert order_change.expires_at == datetime.datetime(
            2020, 1, 17, 10, 42, 14, 545052
        )
        assert order_change.new_total_amount == "121.30"
        assert order_change.new_total_currency == "GBP"
        assert order_change.penalty_total_amount == "15.50"
        assert order_change.penalty_total_currency == "GBP"
        assert order_change.refund_to == "voucher"
        assert order_change.order_id == "ord_0000A3tQcCRZ9R8OY0QlxA"
        assert isinstance(order_change.slices.add, type([]))
        assert isinstance(order_change.slices.remove, type([]))
        assert order_change.slices.add[0].id == "sli_00009htYpSCXrwaB9Dn123"
        assert order_change.slices.remove[0].id == "sli_00009htYpSCXrwaB9Dn123"


def test_create_order_change(requests_mock):
    url = "air/order_changes"
    with fixture("create-order-change", url, requests_mock.post, 201) as client:
        order_change = client.order_changes.create("ord_0000A3tQcCRZ9R8OY0QlxA")

        assert order_change.id == "ocr_0000A3tQSmKyqOrcySrGbo"
        assert order_change.order_id == "ord_0000A3tQcCRZ9R8OY0QlxA"
        assert order_change.confirmed_at is None
        assert order_change.refund_to is None


def test_confirm_order_change(requests_mock):
    url = "air/order_changes/ocr_0000A3tQSmKyqOrcySrGbo/actions/confirm"
    with fixture("confirm-order-change", url, requests_mock.post, 200) as client:
        payment = {"type": "balance", "currency": "GBP", "amount": "30.20"}

        order_change = client.order_changes.confirm(
            "ocr_0000A3tQSmKyqOrcySrGbo", payment
        )

        assert order_change.id == "ocr_0000A3tQSmKyqOrcySrGbo"
        assert order_change.order_id == "ord_0000A3tQcCRZ9R8OY0QlxA"
        assert order_change.confirmed_at == datetime.datetime(2020, 1, 17, 11, 51, 43)
        assert order_change.refund_to == "voucher"


def test_confirm_order_change_with_invalid_data(requests_mock):
    url = "air/order_changes/ocr_0000A3tQSmKyqOrcySrGbo/actions/confirm"
    with fixture("confirm-order-change", url, requests_mock.post, 200) as client:
        with pytest.raises(OrderChangeClient.InvalidPayment):
            client.order_changes.confirm(
                "ocr_0000A3tQSmKyqOrcySrGbo", {"type": "balance", "currency": "GBP"}
            )

        with pytest.raises(OrderChangeClient.InvalidPaymentType):
            client.order_changes.confirm(
                "ocr_0000A3tQSmKyqOrcySrGbo",
                {"type": "unknown", "currency": "GBP", "amount": "30.20"},
            )



from .fixtures import fixture


def test_get_aircraft_by_id(requests_mock):
    url = "air/aircraft/id"
    with fixture("get-aircraft-by-id", url, requests_mock.get, 200) as client:
        aircraft = client.aircraft.get("id")
        assert aircraft.id == "arc_00009UhD4ongolulWd91Ky"
        assert aircraft.name == "Airbus Industries A380"
        assert aircraft.iata_code == "380"


def test_get_aircraft(requests_mock):
    # We need a way to ensure pagination finished in a mocking environment
    end_pagination_url = (
        "http://someaddress/air/aircraft?limit=50"
        + "&after=g2wAAAACbQAAABBBZXJvbWlzdC1LaGFya2l2bQAAAB%3D"
    )
    end_pagination_response = {"meta": {"after": None}, "data": []}
    requests_mock.get(
        end_pagination_url, complete_qs=True, json=end_pagination_response
    )

    url = "air/aircraft?limit=50"
    with fixture("get-aircraft", url, requests_mock.get, 200) as client:
        paginated_aircraft = client.aircraft.list()
        aircraft = list(paginated_aircraft)
        assert len(aircraft) == 1
        aircraft = aircraft[0]
        assert aircraft.id == "arc_00009UhD4ongolulWd91Ky"
        assert aircraft.name == "Airbus Industries A380"
        assert aircraft.iata_code == "380"



from duffel_api.models import OrderCancellation

from .fixtures import raw_fixture


def test_order_cancellation_model_parsing():
    name = "get-order-cancellation-by-id"
    with raw_fixture(name) as fixture:
        assert OrderCancellation.from_json(fixture["data"])



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Sequence

from duffel_api.models import Aircraft, Airline, Place, Airport
from duffel_api.utils import get_and_transform


@dataclass
class OrderChangeOfferSlicesSliceSegment:
    """A segment within a slice that is being removed or added"""

    id: str
    aircraft: Optional[Aircraft]
    arriving_at: datetime
    departing_at: datetime
    destination: Airport
    destination_terminal: Optional[str]
    origin: Airport
    origin_terminal: Optional[str]
    distance: Optional[str]
    duration: Optional[str]
    marketing_carrier: Airline
    marketing_carrier_flight_number: str
    operating_carrier: Airline
    operating_carrier_flight_number: Optional[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            aircraft=get_and_transform(json, "aircraft", Aircraft.from_json),
            arriving_at=datetime.strptime(json["arriving_at"], "%Y-%m-%dT%H:%M:%S"),
            departing_at=datetime.strptime(json["departing_at"], "%Y-%m-%dT%H:%M:%S"),
            destination=Airport.from_json(json["destination"]),
            destination_terminal=json.get("destination_terminal"),
            origin=Airport.from_json(json["origin"]),
            origin_terminal=json.get("origin_terminal"),
            distance=json.get("distance"),
            duration=json.get("duration"),
            marketing_carrier=Airline.from_json(json["marketing_carrier"]),
            marketing_carrier_flight_number=json["marketing_carrier_flight_number"],
            operating_carrier=Airline.from_json(json["operating_carrier"]),
            operating_carrier_flight_number=json.get("operating_carrier_flight_number"),
        )


@dataclass
class OrderChangeOfferSlicesSlice:
    """A slice that is being removed or added"""

    id: str
    duration: Optional[str]
    destination: Place
    destination_type: str
    origin: Place
    origin_type: str
    segments: Sequence[OrderChangeOfferSlicesSliceSegment]

    allowed_place_types = ["airport", "city"]

    class InvalidPlaceType(Exception):
        """Invalid type of place"""

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            destination_type=json["destination_type"],
            destination=Place.from_json(json["destination"]),
            origin_type=json["origin_type"],
            origin=Place.from_json(json["origin"]),
            duration=json.get("duration"),
            segments=get_and_transform(
                json,
                "segments",
                lambda value: [
                    OrderChangeOfferSlicesSliceSegment.from_json(segment)
                    for segment in value
                ],
                [],
            ),
        )


@dataclass
class OrderChangeOfferSlices:
    """The slices to be added and/or removed"""

    add: Sequence[OrderChangeOfferSlicesSlice]
    remove: Sequence[OrderChangeOfferSlicesSlice]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            add=get_and_transform(
                json,
                "add",
                lambda value: [
                    OrderChangeOfferSlicesSlice.from_json(slice) for slice in value
                ],
                [],
            ),
            remove=get_and_transform(
                json,
                "remove",
                lambda value: [
                    OrderChangeOfferSlicesSlice.from_json(slice) for slice in value
                ],
                [],
            ),
        )


@dataclass
class OrderChangeOffer:
    """After you've searched for flights to add to your order by creating an order change
    request, we'll send your search to a range of airlines, which may return order change
    offers.

    Each order change offer represents flights you can buy from an airline at a particular
    price that meets your search criteria.

    You'll see slices inside the offers. Each slice will also include a list of one or
    more specific flights (called segments) that the airline is offering to get the
    passengers where they want to go.
    """

    id: str
    order_change_id: str
    live_mode: bool
    change_total_amount: str
    change_total_currency: Optional[str]
    new_total_amount: Optional[str]
    new_total_currency: str
    penalty_amount: Optional[str]
    penalty_currency: Optional[str]
    refund_to: str
    created_at: datetime
    updated_at: datetime
    expires_at: datetime
    slices: OrderChangeOfferSlices

    allowed_refund_types = [
        "arc_bsp_cash",
        "balance",
        "card",
        "original_form_of_payment",
        "voucher",
        "awaiting_payment",
    ]

    class InvalidRefundType(Exception):
        """Invalid refund type provided"""

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            order_change_id=json["order_change_id"],
            live_mode=json["live_mode"],
            change_total_amount=json["change_total_amount"],
            change_total_currency=json.get("change_total_currency"),
            new_total_amount=json.get("new_total_amount"),
            new_total_currency=json["new_total_currency"],
            penalty_amount=json.get("penalty_amount"),
            penalty_currency=json.get("penalty_currency"),
            refund_to=json["refund_to"],
            created_at=datetime.strptime(json["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            updated_at=datetime.strptime(json["updated_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            expires_at=parse_datetime(json["expires_at"]),
            slices=OrderChangeOfferSlices.from_json(json["slices"]),
        )


def parse_datetime(value: str) -> datetime:
    # There are inconsistent formats used for this field depending on the
    # endpoint
    if len(value) == 20:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ")
    else:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")



from ...http_client import HttpClient
from ...models import Payment


class PaymentClient(HttpClient):
    """Client to interact with Payments."""

    class InvalidPayment(Exception):
        """Invalid payment data provided"""

    class InvalidPaymentType(Exception):
        """Invalid payment type provided"""

    def __init__(self, **kwargs):
        self._url = "/air/payments"
        super().__init__(**kwargs)

    def create(self):
        """Initiate creation of a Payment."""
        return PaymentCreate(self)


class PaymentCreate:
    """Auxiliary class to provide methods for payment creation related data"""

    def __init__(self, client):
        self._client = client
        self._order_id = None
        self._payment = None

    @staticmethod
    def _validate_payment(payment):
        """Validate payment details"""
        if set(payment.keys()) != set(["amount", "currency", "type"]):
            raise PaymentClient.InvalidPayment(payment)
        if payment["type"] not in ["arc_bsp_cash", "balance", "payments"]:
            raise PaymentClient.InvalidPaymentType(payment["type"])

    def order(self, order_id):
        """Add order identifier"""
        self._order_id = order_id
        return self

    def payment(self, payment):
        """Add payment details"""
        self._payment = payment
        return self

    def execute(self):
        """POST /air/payments."""
        PaymentCreate._validate_payment(self._payment)
        res = self._client.do_post(
            self._client._url,
            body={"data": {"order_id": self._order_id, "payment": self._payment}},
        )
        return Payment.from_json(res["data"])



from ...http_client import HttpClient, Pagination
from ...models import OrderCancellation


class OrderCancellationClient(HttpClient):
    """Entrypoint to create to cancel orders.

    To cancel an order, you'll need to create an order cancellation, check the
    refund_amount returned, and, if you're happy to go ahead and cancel the order.

    The refund specified by refund_amount, if any, will be returned to your original
    payment method (i.e. your Duffel balance). You'll then need to refund your customer
    (e.g. back to their credit/debit card).
    """

    def __init__(self, **kwargs):
        """Instantiate an order cancellation client."""
        self._url = "/air/order_cancellations"
        super().__init__(**kwargs)

    def get(self, id_):
        """GET /air/order_cancellations/:id."""
        res = self.do_get(f"{self._url}/{id_}")
        if res is not None:
            return OrderCancellation.from_json(res["data"])

    def list(self, order_id, limit=50):
        """Retrieve a paginated list of order cancellations."""
        params = {"limit": limit, "order_id": order_id}
        return Pagination(self, OrderCancellation, params)

    def create(self, order_id):
        """Create an order cancellation.

        To begin the process of cancelling an order you need to create an order
        cancellation. The OrderCancellation will contain the refund_amount due from the
        Airline.

        To proceed, you must confirm the cancellation using the Confirm an order
        cancellation endpoint.

        """
        res = self.do_post(self._url, body={"data": {"order_id": order_id}})
        if res is not None:
            return OrderCancellation.from_json(res["data"])

    def confirm(self, id_):
        """Confirm an order cancellation.

        Once you've created a pending order cancellation, you'll know the refund_amount
        you're due to get back.

        To actually cancel the order, you'll need to confirm the cancellation. The booking
        with the airline will be cancelled, and the refund_amount will be returned to the
        original payment method (i.e. your Duffel balance). You'll then need to refund
        your customer (e.g. back to their credit/debit card).

        """
        url = f"{self._url}/{id_}/actions/confirm"
        res = self.do_post(url)
        if res is not None:
            return OrderCancellation.from_json(res["data"])



import pytest
from .fixtures import fixture

from duffel_api.api import OfferClient


def test_get_offer_by_id(requests_mock):
    url = "air/offers/id"
    with fixture("get-offer-by-id", url, requests_mock.get, 200) as client:
        offer = client.offers.get("id")
        assert offer.id == "off_00009htYpSCXrwaB9DnUm0"
        assert len(offer.slices) == 1
        assert offer.owner.id == "aln_00001876aqC8c5umZmrRds"
        assert offer.owner.name == "British Airways"
        assert len(offer.passengers) == 1
        assert len(offer.available_services) == 1
        service = offer.available_services[0]
        assert service.type == "baggage"
        assert service.total_amount == "15.00"
        assert service.total_currency == "GBP"


def test_get_offers(requests_mock):
    # We need a way to ensure pagination finished in a mocking environment
    end_pagination_url = (
        "http://someaddress/air/offers?limit=50"
        + "&offer_request_id=offer_request_id"
        + "&after=g2wAAAACbQAAABBBZXJvbWlzdC1LaGFya2l2bQAAAB%3D"
    )
    end_pagination_response = {"meta": {"after": None}, "data": []}
    requests_mock.get(
        end_pagination_url, complete_qs=True, json=end_pagination_response
    )

    url = "air/offers?limit=50&offer_request_id=offer_request_id"
    with fixture("get-offers", url, requests_mock.get, 200) as client:
        paginated_offers = client.offers.list("offer_request_id")
        offers = list(paginated_offers)
        assert len(offers) == 1
        offer = offers[0]
        assert offer.id == "off_00009htYpSCXrwaB9DnUm0"
        assert len(offer.slices) == 2
        slice_1 = offer.slices[0]
        assert slice_1.conditions.change_before_departure.allowed is True
        assert slice_1.conditions.change_before_departure.penalty_amount == "100.00"
        assert slice_1.conditions.change_before_departure.penalty_currency == "GBP"
        assert len(slice_1.segments) == 1
        assert slice_1.segments[0].id == "seg_00009htYpSCXrwaB9Dn456"
        slice_2 = offer.slices[1]
        assert slice_2.conditions.change_before_departure is None
        assert len(slice_2.segments) == 1
        assert slice_2.segments[0].id == "seg_0000AEfTff29Oo8LXf7FYG"


def test_offer_update_passenger(requests_mock):
    offer_id = "off_00009htYpSCXrwaB9DnUm0"
    offer_passenger_id = "pas_00009hj8USM7Ncg31cBCL"

    url = f"air/offers/{offer_id}/passengers/{offer_passenger_id}"

    with fixture(
        "update-offer-passenger-by-id", url, requests_mock.patch, 200
    ) as client:
        offer_passenger = client.offers.update_passenger(
            offer_id,
            offer_passenger_id,
            "Earhart",
            "Amelia",
            [{"account_number": "12901014", "airline_iata_code": "BA"}],
        )

        assert offer_passenger.id == offer_passenger_id
        assert offer_passenger.family_name == "Earhart"
        assert offer_passenger.given_name == "Amelia"
        assert (
            offer_passenger.loyalty_programme_accounts[0].account_number == "12901014"
        )
        assert offer_passenger.loyalty_programme_accounts[0].airline_iata_code == "BA"


def test_offer_update_passenger_with_invalid_data(requests_mock):
    offer_id = "off_00009htYpSCXrwaB9DnUm0"
    offer_passenger_id = "pas_00009hj8USM7Ncg31cBCL"

    url = f"air/offers/{offer_id}/passengers/{offer_passenger_id}"

    with fixture(
        "update-offer-passenger-by-id", url, requests_mock.patch, 200
    ) as client:
        with pytest.raises(OfferClient.InvalidOfferId):
            client.offers.update_passenger(
                "",
                offer_passenger_id,
                "Earhart",
                "Amelia",
                [{"account_number": "12901014", "airline_iata_code": "BA"}],
            )

        with pytest.raises(OfferClient.InvalidOfferPassengerId):
            client.offers.update_passenger(
                offer_id,
                "",
                "Earhart",
                "Amelia",
                [{"account_number": "12901014", "airline_iata_code": "BA"}],
            )

        with pytest.raises(OfferClient.InvalidFamilyName):
            client.offers.update_passenger(
                offer_id,
                offer_passenger_id,
                "",
                "Amelia",
                [{"account_number": "12901014", "airline_iata_code": "BA"}],
            )

        with pytest.raises(OfferClient.InvalidGivenName):
            client.offers.update_passenger(
                offer_id,
                offer_passenger_id,
                "Earhart",
                "",
                [{"account_number": "12901014", "airline_iata_code": "BA"}],
            )

        with pytest.raises(OfferClient.InvalidLoyaltyProgrammeAirlineIataCode):
            client.offers.update_passenger(
                offer_id,
                offer_passenger_id,
                "Earhart",
                "Amelia",
                [{"account_number": "12901014", "airline_iata_code": ""}],
            )

        with pytest.raises(OfferClient.InvalidLoyaltyProgrammeAccountNumber):
            client.offers.update_passenger(
                offer_id,
                offer_passenger_id,
                "Earhart",
                "Amelia",
                [{"account_number": "", "airline_iata_code": "BA"}],
            )



from ...http_client import HttpClient
from ...models import OrderChangeRequest


class OrderChangeRequestClient(HttpClient):
    """Client to interact with Order Change Requests"""

    def __init__(self, **kwargs):
        """Instantiate an order change request client."""
        self._url = "/air/order_change_requests"
        super().__init__(**kwargs)

    def create(self, order_id):
        """POST /air/order_change_requests"""
        return OrderChangeRequestCreate(self, order_id)

    def get(self, id_):
        """GET /air/order_change_requests/:id"""
        res = self.do_get(f"{self._url}/{id_}")
        if res is not None:
            return OrderChangeRequest.from_json(res["data"])


class OrderChangeRequestCreate:
    """Auxiliary class for order change request creation related data"""

    class InvalidOrderId(Exception):
        """Invalid order ID provided"""

    class InvalidSlices(Exception):
        """Invalid slices data provided"""

    def __init__(self, client, order_id):
        """Instantiate an order change request creation."""
        self._client = client
        self._order_id = order_id
        self._slices = []
        OrderChangeRequestCreate._validate_order_id(self._order_id)

    @staticmethod
    def _validate_order_id(order_id):
        """Set order ID"""
        if type(order_id) is not str:
            raise OrderChangeRequestCreate.InvalidOrderId(order_id)

    @staticmethod
    def _validate_slices(slices):
        """Validate structure of Metadata"""
        if type(slices) is not dict:
            raise OrderChangeRequestCreate.InvalidSlices(slices)

    def slices(self, slices):
        """Set slices"""
        self._slices = slices
        OrderChangeRequestCreate._validate_slices(self._slices)
        return self

    def execute(self):
        """POST /air/order_change_requests - trigger the call to create the
        order change request"""
        OrderChangeRequestCreate._validate_slices(self._slices)
        res = self._client.do_post(
            self._client._url,
            body={
                "data": {
                    "order_id": self._order_id,
                    "slices": self._slices,
                }
            },
        )
        return OrderChangeRequest.from_json(res["data"])



from .fixtures import fixture


def test_get_airport_by_id(requests_mock):
    url = "air/airports/id"
    with fixture("get-airport-by-id", url, requests_mock.get, 200) as client:
        airport = client.airports.get("id")
        assert airport.id == "arp_lhr_gb"
        assert airport.name == "Heathrow"
        assert airport.iata_code == "LHR"
        assert airport.city.name == "London"


def test_get_airports(requests_mock):
    # We need a way to ensure pagination finished in a mocking environment
    end_pagination_url = (
        "http://someaddress/air/airports?limit=50"
        + "&after=g2wAAAACbQAAABBBZXJvbWlzdC1LaGFya2l2bQAAAB%3D"
    )
    end_pagination_response = {"meta": {"after": None}, "data": []}
    requests_mock.get(
        end_pagination_url, complete_qs=True, json=end_pagination_response
    )

    url = "air/airports?limit=50"
    with fixture("get-airports", url, requests_mock.get, 200) as client:
        paginated_airports = client.airports.list()
        airports = list(paginated_airports)
        assert len(airports) == 1
        airport = airports[0]
        assert airport.id == "arp_lhr_gb"
        assert airport.name == "Heathrow"
        assert airport.iata_code == "LHR"
        assert airport.city.name == "London"



from dataclasses import dataclass
from datetime import date, datetime
from typing import Sequence

from duffel_api.models import OrderChangeOffer
from duffel_api.utils import get_and_transform


@dataclass
class OrderChangeRequestSliceAdd:
    """The slice to be added"""

    cabin_class: str
    departure_date: date
    destination: str
    origin: str

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            cabin_class=json["cabin_class"],
            departure_date=date.fromisoformat(json["departure_date"]),
            destination=json["destination"],
            origin=json["origin"],
        )


@dataclass
class OrderChangeRequestSliceRemove:
    """The slice to be removed"""

    slice_id: str

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            slice_id=json["slice_id"],
        )


@dataclass
class OrderChangeRequestSlices:
    """The slices to be added and/or removed"""

    add: Sequence[OrderChangeRequestSliceAdd]
    remove: Sequence[OrderChangeRequestSliceRemove]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            add=get_and_transform(
                json,
                "add",
                lambda value: [
                    OrderChangeRequestSliceAdd.from_json(slice) for slice in value
                ],
                [],
            ),
            remove=get_and_transform(
                json,
                "remove",
                lambda value: [
                    OrderChangeRequestSliceRemove.from_json(slice) for slice in value
                ],
                [],
            ),
        )


@dataclass
class OrderChangeRequest:
    """To change an order, you'll need to create an order change request. An
    order change request describes the slices of an existing paid order that you
    want to remove and search criteria for new slices you want to add.
    """

    id: str
    live_mode: bool
    created_at: datetime
    updated_at: datetime
    order_id: str
    slices: OrderChangeRequestSlices
    order_change_offers: Sequence[OrderChangeOffer]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            live_mode=json["live_mode"],
            order_id=json["order_id"],
            created_at=datetime.strptime(json["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            updated_at=datetime.strptime(json["updated_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            slices=OrderChangeRequestSlices.from_json(json["slices"]),
            order_change_offers=get_and_transform(
                json,
                "order_change_offers",
                lambda value: [
                    OrderChangeOffer.from_json(order_change_offer)
                    for order_change_offer in value
                ],
                [],
            ),
        )



from ...http_client import HttpClient, Pagination
from ...models import OrderChangeOffer


class OrderChangeOfferClient(HttpClient):
    """Client to interact with Order Change Offers"""

    def __init__(self, **kwargs):
        self._url = "/air/order_change_offers"
        super().__init__(**kwargs)

    def get(self, id_):
        """GET /air/order_change_offers/:id"""
        res = self.do_get(f"{self._url}/{id_}")
        if res is not None:
            return OrderChangeOffer.from_json(res["data"])

    def list(self, order_change_request_id, sort=None, max_connections=None, limit=50):
        """GET /air/order_change_offers"""
        params = {"limit": limit, "order_change_request_id": order_change_request_id}
        if sort:
            params["sort"] = sort
        if max_connections:
            params["max_connections"] = max_connections
        return Pagination(self, OrderChangeOffer, params)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Sequence, Union

from duffel_api.models import Aircraft, Airline, Airport, LoyaltyProgrammeAccount, Place
from duffel_api.utils import get_and_transform


@dataclass
class OrderConditionChangeBeforeDeparture:
    """Whether the whole order can be changed before the departure of the first slice.

    If all of the slices on the order can be changed then
    the `allowed` property will be `true`. Refer to the `slices` for
    information about change penalties.

    If any of the slices on the order can't be changed then
    the `allowed` property will be `false`. In this case you should
    refer to the `slices` conditions to determine if any part
    of the order is changeable.

    If the airline hasn't provided any information about whether
    this order can be changed then this property will be `null`.
    """

    allowed: bool
    penalty_amount: Optional[str]
    penalty_currency: Optional[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            allowed=json["allowed"],
            penalty_amount=json.get("penalty_amount"),
            penalty_currency=json.get("penalty_currency"),
        )


@dataclass
class OrderConditionRefundBeforeDeparture:
    """Whether the whole order can be refunded before the departure of the first slice.

    If all of the slices on the order can be refunded then
    the `allowed` property will be `true` and information will be
    provided about any penalties.

    If any of the slices on the order can't be refunded then
    the `allowed` property will be `false`.

    If the airline hasn't provided any information about whether
    this order can be refunded then this property will be `null`.
    """

    allowed: bool
    penalty_amount: Optional[str]
    penalty_currency: Optional[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            allowed=json["allowed"],
            penalty_amount=json.get("penalty_amount"),
            penalty_currency=json.get("penalty_currency"),
        )


@dataclass
class OrderConditions:
    """The conditions associated with this order, describing the kinds of
    modifications you can make to it and any penalties that
    will apply to those modifications.

    This information assumes the condition is applied to all of the slices and passengers
    associated with this order - for information at the slice level
    (e.g. "what happens if I just want to change the first slice?") refer to the `slices`.

    If a particular kind of modification is allowed, you may not always be able to take
    action through the Duffel API. In some cases, you may need to contact the Duffel
    support team or the airline directly.
    """

    change_before_departure: Optional[OrderConditionChangeBeforeDeparture]
    refund_before_departure: Optional[OrderConditionRefundBeforeDeparture]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            change_before_departure=get_and_transform(
                json,
                "change_before_departure",
                OrderConditionChangeBeforeDeparture.from_json,
            ),
            refund_before_departure=get_and_transform(
                json,
                "refund_before_departure",
                OrderConditionRefundBeforeDeparture.from_json,
            ),
        )


@dataclass
class OrderSliceSegmentPassengerSeat:
    """An object containing metadata about the service, like the designator of the seat"""

    designator: str
    name: str
    disclosures: Sequence[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            designator=json["designator"],
            name=json["name"],
            disclosures=json["disclosures"],
        )


@dataclass
class OrderSliceSegmentPassengerBaggage:
    """The baggage allowances for the passenger on this segment that were included in the
    original offer. Any extra baggage items which were booked as services will be listed
    in the services field instead of here.
    """

    type: str
    quantity: int

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            type=json["type"],
            quantity=json["quantity"],
        )


@dataclass
class OrderSliceSegmentPassenger:
    """Additional segment-specific information about the passengers included in the offer
    (e.g. their baggage allowance and the cabin class they will be travelling in)
    """

    baggages: Sequence[OrderSliceSegmentPassengerBaggage]
    cabin_class: str
    cabin_class_marketing_name: str
    passenger_id: str
    seat: Optional[OrderSliceSegmentPassengerSeat]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            baggages=get_and_transform(
                json,
                "baggages",
                lambda value: [
                    OrderSliceSegmentPassengerBaggage.from_json(baggage)
                    for baggage in value
                ],
                [],
            ),
            cabin_class=json["cabin_class"],
            cabin_class_marketing_name=json["cabin_class_marketing_name"],
            passenger_id=json["passenger_id"],
            seat=get_and_transform(
                json, "seat", OrderSliceSegmentPassengerSeat.from_json
            ),
        )


@dataclass
class OrderSliceSegment:
    """The segments - that is, specific flights - that the airline is offering to get the
    passengers from the `origin` to the `destination`
    """

    id: str
    aircraft: Optional[Aircraft]
    arriving_at: datetime
    departing_at: datetime
    destination: Airport
    destination_terminal: Optional[str]
    origin: Airport
    origin_terminal: Optional[str]
    distance: Optional[str]
    duration: Optional[str]
    marketing_carrier: Airline
    marketing_carrier_flight_number: str
    operating_carrier: Airline
    operating_carrier_flight_number: Optional[str]
    passengers: Sequence[OrderSliceSegmentPassenger]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            aircraft=get_and_transform(json, "aircraft", Aircraft.from_json),
            arriving_at=datetime.strptime(json["arriving_at"], "%Y-%m-%dT%H:%M:%S"),
            departing_at=datetime.strptime(json["departing_at"], "%Y-%m-%dT%H:%M:%S"),
            destination=Airport.from_json(json["destination"]),
            destination_terminal=json.get("destination_terminal"),
            origin=Airport.from_json(json["origin"]),
            origin_terminal=json.get("origin_terminal"),
            distance=json.get("distance"),
            duration=json.get("duration"),
            marketing_carrier=Airline.from_json(json["marketing_carrier"]),
            marketing_carrier_flight_number=json["marketing_carrier_flight_number"],
            operating_carrier=Airline.from_json(json["operating_carrier"]),
            operating_carrier_flight_number=json.get("operating_carrier_flight_number"),
            passengers=get_and_transform(
                json,
                "passengers",
                lambda value: [
                    OrderSliceSegmentPassenger.from_json(passenger)
                    for passenger in value
                ],
                [],
            ),
        )


@dataclass
class OrderSliceConditionChangeBeforeDeparture:
    """Whether this slice can be changed before the departure.

    If the slice can be changed for all of the passengers then
    the `allowed` property will be `true` and information will be
    provided about any penalties.

    If none of the passengers on the slice can be changed then
    the `allowed` property will be `false`.

    In all other cases this property will be `null` indicating we
    can't provide the information for this slice.

    The `penalty_amount` is specific to changing this slice and may not be
    the penalties that is applied if the entire order is changed.
    """

    allowed: bool
    penalty_amount: Optional[str]
    penalty_currency: Optional[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            allowed=json["allowed"],
            penalty_amount=json.get("penalty_amount"),
            penalty_currency=json.get("penalty_currency"),
        )


@dataclass
class OrderSliceConditions:
    """The conditions associated with this slice, describing the kinds of
    modifications you can make and any penalties that
    will apply to those modifications.

    This condition is applied only to this slice and to all the passengers
    associated with this order - for information at the order level
    (e.g. "what happens if I want to change all the slices?") refer to the
    `conditions` at the top level.

    If a particular kind of modification is allowed, you may not always be able to take
    action through the Duffel API. In some cases, you may need to contact the Duffel
    support team or
    the airline directly.
    """

    change_before_departure: Optional[OrderSliceConditionChangeBeforeDeparture]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            change_before_departure=get_and_transform(
                json,
                "change_before_departure",
                OrderSliceConditionChangeBeforeDeparture.from_json,
            )
        )


@dataclass
class OrderSlice:
    """A slice is one part of the slices that make up the itinerary of an order.
    One-way journeys can be expressed using one slice, whereas return trips will need two.
    """

    id: str
    changeable: bool
    destination_type: str
    destination: Place
    origin_type: str
    origin: Place
    duration: Optional[str]
    segments: Sequence[OrderSliceSegment]
    conditions: OrderSliceConditions

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            changeable=json["changeable"],
            destination_type=json["destination_type"],
            destination=Place.from_json(json["destination"]),
            origin_type=json["origin_type"],
            origin=Place.from_json(json["origin"]),
            duration=json.get("duration"),
            segments=get_and_transform(
                json,
                "segments",
                lambda value: [
                    OrderSliceSegment.from_json(segment) for segment in value
                ],
                [],
            ),
            conditions=OrderSliceConditions.from_json(json["conditions"]),
        )


@dataclass
class OrderServiceMetadataSeat:
    """An object containing metadata about the service, like the designator of the seat"""

    designator: str
    name: str
    disclosures: Sequence[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            designator=json["designator"],
            name=json["name"],
            disclosures=json["disclosures"],
        )


@dataclass
class OrderServiceMetadataBaggage:
    """An object containing metadata about the service, like the maximum weight and
    dimensions of the baggage.
    """

    type: str
    maximum_weight_kg: Optional[int]
    maximum_height_cm: Optional[int]
    maximum_length_cm: Optional[int]
    maximum_depth_cm: Optional[int]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            type=json["type"],
            maximum_weight_kg=json.get("maximum_weight_kg"),
            maximum_height_cm=json.get("maximum_height_cm"),
            maximum_length_cm=json.get("maximum_length_cm"),
            maximum_depth_cm=json.get("maximum_depth_cm"),
        )


@dataclass
class OrderService:
    """The service booked along with this order"""

    id: str
    metadata: Union[OrderServiceMetadataSeat, OrderServiceMetadataBaggage]
    passenger_ids: Sequence[str]
    quantity: int
    segment_ids: Sequence[str]
    total_amount: str
    total_currency: str
    type: str

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        type = json["type"]
        metadata = json["metadata"]

        if type == "seat":
            metadata = OrderServiceMetadataSeat.from_json(metadata)
        elif type == "baggage":
            metadata = OrderServiceMetadataBaggage.from_json(metadata)

        return cls(
            id=json["id"],
            metadata=metadata,
            passenger_ids=json["passenger_ids"],
            quantity=json["quantity"],
            segment_ids=json["segment_ids"],
            total_amount=json["total_amount"],
            total_currency=json["total_currency"],
            type=type,
        )


@dataclass
class OrderPaymentStatus:
    """The payment status for an order"""

    awaiting_payment: bool
    payment_required_by: Optional[datetime]
    price_guarantee_expires_at: Optional[datetime]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            awaiting_payment=json["awaiting_payment"],
            payment_required_by=get_and_transform(
                json,
                "payment_required_by",
                lambda value: datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ"),
            ),
            price_guarantee_expires_at=get_and_transform(
                json,
                "price_guarantee_expires_at",
                lambda value: datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ"),
            ),
        )


@dataclass
class OrderPassenger:
    """A passenger who is travelling"""

    id: str
    type: str
    infant_passenger_id: Optional[str]
    given_name: str
    family_name: str
    gender: str
    title: str
    born_on: str
    loyalty_programme_accounts: Sequence[LoyaltyProgrammeAccount]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            type=json["type"],
            infant_passenger_id=json.get("infant_passenger_id"),
            given_name=json["given_name"],
            family_name=json["family_name"],
            gender=json["gender"],
            title=json["title"],
            born_on=json["born_on"],
            loyalty_programme_accounts=get_and_transform(
                json,
                "loyalty_programme_accounts",
                lambda value: [
                    LoyaltyProgrammeAccount.from_json(loyalty_programme_account)
                    for loyalty_programme_account in value
                ],
                [],
            ),
        )


@dataclass
class OrderDocument:
    """A document issued for this order."""

    type: str
    unique_identifier: str

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            type=json["type"],
            unique_identifier=json["unique_identifier"],
        )


@dataclass
class Order:
    """Once you've searched for flights by creating an offer request, and you've chosen
    which offer you want to book, you'll then want to create an order.

    To create an order, you just need to provide the offer ID, payment details and some
    additional information on the passengers (e.g. their full name and date of birth).
    """

    id: str
    live_mode: bool
    available_actions: Sequence[str]
    base_amount: Optional[str]
    base_currency: Optional[str]
    booking_reference: str
    cancelled_at: Optional[datetime]
    content: str
    created_at: datetime
    synced_at: Optional[datetime]
    documents: Sequence[OrderDocument]
    passengers: Sequence[OrderPassenger]
    payment_status: OrderPaymentStatus
    services: Sequence[OrderService]
    slices: Sequence[OrderSlice]
    conditions: OrderConditions
    owner: Airline
    tax_amount: str
    tax_currency: Optional[str]
    total_amount: str
    total_currency: str
    metadata: dict

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            available_actions=json["available_actions"],
            live_mode=json["live_mode"],
            base_amount=json.get("base_amount"),
            base_currency=json.get("base_currency"),
            booking_reference=json["booking_reference"],
            cancelled_at=get_and_transform(
                json,
                "cancelled_at",
                lambda value: datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ"),
            ),
            content=json["content"],
            created_at=datetime.strptime(json["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            synced_at=get_and_transform(
                json,
                "synced_at",
                lambda value: datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ"),
            ),
            documents=get_and_transform(
                json,
                "documents",
                lambda value: [OrderDocument.from_json(document) for document in value],
                [],
            ),
            owner=Airline.from_json(json["owner"]),
            passengers=get_and_transform(
                json,
                "passengers",
                lambda value: [
                    OrderPassenger.from_json(passenger) for passenger in value
                ],
                [],
            ),
            payment_status=OrderPaymentStatus.from_json(json["payment_status"]),
            services=get_and_transform(
                json,
                "services",
                lambda value: [OrderService.from_json(service) for service in value],
                [],
            ),
            slices=get_and_transform(
                json,
                "slices",
                lambda value: [OrderSlice.from_json(slice) for slice in value],
                [],
            ),
            conditions=OrderConditions.from_json(json["conditions"]),
            tax_amount=json["tax_amount"],
            tax_currency=json.get("tax_currency"),
            total_amount=json["total_amount"],
            total_currency=json["total_currency"],
            # Metadata is customer-specified data, so we don't try and parse it
            metadata=json["metadata"],
        )



from ...http_client import HttpClient
from ...models import SeatMap


class SeatMapClient(HttpClient):
    """Client to interact with Seat Maps"""

    def __init__(self, **kwargs):
        self._url = "/air/seat_maps"
        super().__init__(**kwargs)

    def get(self, offer_id):
        """GET /air/seat_maps"""
        res = self.do_get(self._url, query_params={"offer_id": offer_id})
        if res is not None:
            return [SeatMap.from_json(m) for m in res["data"]]



from duffel_api.models import Webhook


def test_webhook_model_parsing():
    json = {
        "url": "https://www.example.com:4000/webhooks",
        "updated_at": "2020-04-11T15:48:11.642Z",
        "secret": "QKfUULLQh+8SegYmIsF6kA==",
        "live_mode": "true",
        "id": "sev_0000A3tQSmKyqOrcySrGbo",
        "events": ["order.created", "order.updated"],
        "created_at": "2020-04-11T15:48:11.642Z",
        "active": "true",
    }
    assert Webhook.from_json(json)



from ...http_client import HttpClient, Pagination
from ...models import OfferRequest


class OfferRequestClient(HttpClient):
    """To search for flights, you'll need to create an offer request.
    An offer request describes the passengers and where and when they want to
     travel (in the form of a list of slices). It may also include additional
    filters (e.g. a particular cabin to travel in).
    """

    def __init__(self, **kwargs):
        self._url = "/air/offer_requests"
        super().__init__(**kwargs)

    def get(self, id_):
        """GET /air/offer_requests/:id"""
        response = self.do_get(f"{self._url}/{id_}")

        if response is not None:
            return OfferRequest.from_json(response["data"])

    def list(self, limit=50):
        """GET /air/offer_requests"""
        return Pagination(self, OfferRequest, {"limit": limit})

    def create(self):
        """Initiate creation of an Offer Request"""
        return OfferRequestCreate(self)


class OfferRequestCreate(object):
    """Auxiliary class to provide methods for offer request creation related data"""

    class InvalidCabinClass(Exception):
        """Invalid cabin class provided"""

    class InvalidNumberOfPassengers(Exception):
        """Invalid number of passengers provided"""

    class InvalidNumberOfSlices(Exception):
        """Invalid number of slices provided"""

    class InvalidPassenger(Exception):
        """Invalid passenger data provided"""

    class InvalidSlice(Exception):
        """Invalid slice data provided"""

    class InvalidMaxConnectionValue(Exception):
        """Invalid max connection value provided"""

    def __init__(self, client):
        self._client = client
        self._return_offers = "false"
        self._cabin_class = "economy"
        self._passengers = []
        self._slices = []
        self._max_connections = 1

    @staticmethod
    def _validate_cabin_class(cabin_class):
        """Validate cabin class"""
        if cabin_class not in [
            "first",
            "business",
            "economy",
            "premium_economy",
        ]:
            raise OfferRequestCreate.InvalidCabinClass(cabin_class)

    @staticmethod
    def _validate_passengers(passengers):
        """Validate passenger count and the data provided for each if any were given"""
        if len(passengers) == 0:
            raise OfferRequestCreate.InvalidNumberOfPassengers(passengers)
        for passenger in passengers:
            if not ("type" in passenger or "age" in passenger):
                raise OfferRequestCreate.InvalidPassenger(passenger)

    @staticmethod
    def _validate_slices(slices):
        """Validate number of slices and the data provided for each if any were given"""
        if len(slices) == 0:
            raise OfferRequestCreate.InvalidNumberOfSlices(slices)
        for travel_slice in slices:
            if set(travel_slice.keys()) != set(
                ["departure_date", "destination", "origin"]
            ):
                raise OfferRequestCreate.InvalidSlice(travel_slice)

    @staticmethod
    def _validate_max_connections(max_connections):
        """Validate the max connection number"""
        if not isinstance(max_connections, int) or max_connections < 0:
            raise OfferRequestCreate.InvalidMaxConnectionValue(max_connections)

    def return_offers(self):
        """Set return_offers to 'true'"""
        self._return_offers = "true"
        return self

    def cabin_class(self, cabin_class):
        """Set cabin_class - defaults to 'economy'"""
        OfferRequestCreate._validate_cabin_class(cabin_class)
        self._cabin_class = cabin_class
        return self

    def passengers(self, passengers):
        """Set the passengers that will be travelling"""
        OfferRequestCreate._validate_passengers(passengers)
        self._passengers = passengers
        return self

    def slices(self, slices):
        """Set the slices for the origin-destination we want to travel"""
        OfferRequestCreate._validate_slices(slices)
        self._slices = slices
        return self

    def max_connections(self, max_connections):
        """Set the max_connections for the journey we want to travel"""
        OfferRequestCreate._validate_max_connections(max_connections)
        self._max_connections = max_connections
        return self

    def execute(self):
        """POST /air/offer_requests - trigger the call to create the offer_request"""
        OfferRequestCreate._validate_passengers(self._passengers)
        OfferRequestCreate._validate_slices(self._slices)
        res = self._client.do_post(
            self._client._url,
            query_params={"return_offers": self._return_offers},
            body={
                "data": {
                    "cabin_class": self._cabin_class,
                    "passengers": self._passengers,
                    "max_connections": self._max_connections,
                    "slices": self._slices,
                }
            },
        )
        return OfferRequest.from_json(res["data"])



from duffel_api.models import PaymentIntent

from .fixtures import raw_fixture


def test_payment_intent_model_parsing():
    name = "get-payment-intent-by-id"
    with raw_fixture(name) as fixture:
        assert PaymentIntent.from_json(fixture["data"])



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Sequence

from duffel_api.models import Aircraft, Airline, Airport, LoyaltyProgrammeAccount, Place
from duffel_api.utils import get_and_transform


@dataclass
class OfferConditionChangeBeforeDeparture:
    """Whether the whole offer can be changed before the departure of the first slice.

    If all of the slices on the offer can be changed then the allowed property will be
    true. Refer to the slices for information about change penalties. If any of the slices
    on the offer can't be changed then the allowed property will be false. In this case
    you should refer to the slices conditions to determine if any part of the offer is
    changeable. If the airline hasn't provided any information about whether this offer
    can be changed then this property will be null.
    """

    allowed: bool
    penalty_amount: Optional[str]
    penalty_currency: Optional[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            allowed=json["allowed"],
            penalty_amount=json.get("penalty_amount"),
            penalty_currency=json.get("penalty_currency"),
        )


@dataclass
class OfferConditionRefundBeforeDeparture:
    """Whether the whole offer can be refunded before the departure of the first slice.

    If all of the slices on the offer can be refunded then the allowed property will be
    true and information will be provided about any penalties. If any of the slices on the
    offer can't be refunded then the allowed property will be false. If the airline hasn't
    provided any information about whether this offer can be refunded then this property
    will be null.
    """

    allowed: bool
    penalty_amount: Optional[str]
    penalty_currency: Optional[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            allowed=json["allowed"],
            penalty_amount=json.get("penalty_amount"),
            penalty_currency=json.get("penalty_currency"),
        )


@dataclass
class OfferConditions:
    """The conditions associated with this offer, describing the kinds of modifications
    you can make post-booking and any penalties that will apply to those modifications.
    """

    change_before_departure: Optional[OfferConditionChangeBeforeDeparture]
    refund_before_departure: Optional[OfferConditionRefundBeforeDeparture]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            change_before_departure=get_and_transform(
                json,
                "change_before_departure",
                OfferConditionChangeBeforeDeparture.from_json,
            ),
            refund_before_departure=get_and_transform(
                json,
                "refund_before_departure",
                OfferConditionRefundBeforeDeparture.from_json,
            ),
        )


@dataclass
class PaymentRequirements:
    """The payment requirements for an offer"""

    payment_required_by: Optional[datetime]
    price_guarantee_expires_at: Optional[datetime]
    requires_instant_payment: bool

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            payment_required_by=get_and_transform(
                json,
                "payment_required_by",
                lambda value: datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ"),
            ),
            price_guarantee_expires_at=get_and_transform(
                json,
                "price_guarantee_expires_at",
                lambda value: datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ"),
            ),
            requires_instant_payment=json["requires_instant_payment"],
        )


@dataclass
class ServiceMetadata:
    """An object containing metadata about the service, like the maximum weight
    and dimensions of the baggage.
    """

    # Baggage
    type: Optional[str]
    maximum_weight_kg: Optional[int]
    maximum_height_cm: Optional[int]
    maximum_length_cm: Optional[int]
    maximum_depth_cm: Optional[int]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            type=json.get("type"),
            maximum_weight_kg=json.get("maximum_weight_kg"),
            maximum_height_cm=json.get("maximum_height_cm"),
            maximum_length_cm=json.get("maximum_length_cm"),
            maximum_depth_cm=json.get("maximum_depth_cm"),
        )


@dataclass
class Service:
    """The services that can be booked with the offer but are not included by default,
    for example an additional checked bag. This field is only returned in the [Get single
    offer](https://duffel.com/docs/api/offers/get-offer-by-id) endpoint. When there are no
    services available, or we don't support services for the airline, this list will be
    empty. If you want to know which airlines we support services for, please get in touch
    with the Duffel support team at help@duffel.com.
    """

    id: str
    maximum_quantity: int
    metadata: ServiceMetadata
    passenger_ids: Sequence[str]
    segment_ids: Sequence[str]
    total_amount: str
    total_currency: str
    type: str

    allowed_types = ["baggage"]

    class InvalidType(Exception):
        """Invalid service type"""

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            maximum_quantity=json["maximum_quantity"],
            metadata=ServiceMetadata.from_json(json["metadata"]),
            passenger_ids=json["passenger_ids"],
            segment_ids=json["segment_ids"],
            total_amount=json["total_amount"],
            total_currency=json["total_currency"],
            type=json["type"],
        )


@dataclass
class OfferSliceSegmentPassengerBaggage:
    """The baggage allowances for the passenger on this segment included in the offer.
    Some airlines may allow additional baggage to be booked as a service - see the offer's
    `available_services`
    """

    type: str
    quantity: int

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            type=json["type"],
            quantity=json["quantity"],
        )


@dataclass
class OfferSliceSegmentPassenger:
    """Additional segment-specific information about the passengers included in the offer
    (e.g. their baggage allowance and the cabin class they will be travelling in)

    """

    baggages: Sequence[OfferSliceSegmentPassengerBaggage]
    cabin_class: str
    cabin_class_marketing_name: str
    passenger_id: str
    fare_basis_code: Optional[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            baggages=get_and_transform(
                json,
                "baggages",
                lambda value: [
                    OfferSliceSegmentPassengerBaggage.from_json(baggage)
                    for baggage in value
                ],
                [],
            ),
            cabin_class=json["cabin_class"],
            cabin_class_marketing_name=json["cabin_class_marketing_name"],
            passenger_id=json["passenger_id"],
            fare_basis_code=json.get("fare"),
        )


@dataclass
class OfferSliceSegment:
    """The segments - that is, specific flights - that the airline is offering
    to get the passengers from the `origin` to the `destination`
    """

    id: str
    aircraft: Optional[Aircraft]
    arriving_at: datetime
    departing_at: datetime
    destination: Airport
    destination_terminal: Optional[str]
    origin: Airport
    origin_terminal: Optional[str]
    distance: Optional[str]
    duration: Optional[str]
    marketing_carrier: Airline
    marketing_carrier_flight_number: str
    operating_carrier: Airline
    operating_carrier_flight_number: Optional[str]
    passengers: Sequence[OfferSliceSegmentPassenger]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            aircraft=get_and_transform(json, "aircraft", Aircraft.from_json),
            arriving_at=datetime.strptime(json["arriving_at"], "%Y-%m-%dT%H:%M:%S"),
            departing_at=datetime.strptime(json["departing_at"], "%Y-%m-%dT%H:%M:%S"),
            destination=Airport.from_json(json["destination"]),
            destination_terminal=json.get("destination_terminal"),
            origin=Airport.from_json(json["origin"]),
            origin_terminal=json.get("origin_terminal"),
            distance=json.get("distance"),
            duration=json.get("duration"),
            marketing_carrier=Airline.from_json(json["marketing_carrier"]),
            marketing_carrier_flight_number=json["marketing_carrier_flight_number"],
            operating_carrier=Airline.from_json(json["operating_carrier"]),
            operating_carrier_flight_number=json.get("operating_carrier_flight_number"),
            passengers=get_and_transform(
                json,
                "passengers",
                lambda value: [
                    OfferSliceSegmentPassenger.from_json(passenger)
                    for passenger in value
                ],
                [],
            ),
        )


@dataclass
class OfferSliceConditionsChangeBeforeDeparture:
    """Whether this slice can be changed before the departure.

    If the slice can be changed for all of the passengers then
    the `allowed` property will be `true` and information will be
    provided about any penalties.

    If none of the passengers on the slice can be changed then
    the `allowed` property will be `false`.

    In all other cases this property will be `null` indicating we
    can't provide the information for this slice.

    The `penalty_amount` is specific to changing this slice and may not be
    the penalty that is applied if the entire offer is changed.
    """

    allowed: bool
    penalty_amount: Optional[str]
    penalty_currency: Optional[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            allowed=json["allowed"],
            penalty_amount=json.get("penalty_amount"),
            penalty_currency=json.get("penalty_currency"),
        )


@dataclass
class OfferSliceConditions:
    """The conditions associated with this slice, describing the kinds of
    modifications you can make post-booking and any penalties that
    will apply to those modifications.

    This condition is applied only to this slice and to all the passengers
    associated with this offer - for information at the offer level
    (e.g. "what happens if I want to change all the slices?") refer to the
    `conditions` at the top level.

    If a particular kind of modification is allowed, you may not always be able to take
    action through the Duffel API. In some cases, you may need to contact the Duffel
    support team or the airline directly.
    """

    change_before_departure: Optional[OfferSliceConditionsChangeBeforeDeparture]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            change_before_departure=get_and_transform(
                json,
                "change_before_departure",
                OfferSliceConditionsChangeBeforeDeparture.from_json,
            ),
        )


@dataclass
class OfferSlice:
    """Each slice will include one or more segments, the specific flights that the airline
    is offering to take the passengers from the slice's origin to its destination.
    """

    id: str
    destination: Place
    destination_terminal: Optional[str]
    origin: Place
    origin_terminal: Optional[str]
    duration: Optional[str]
    fare_brand_name: Optional[str]
    segments: Sequence[OfferSliceSegment]
    conditions: OfferSliceConditions

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            destination=Place.from_json(json["destination"]),
            destination_terminal=json.get("destination_terminal"),
            origin=Place.from_json(json["origin"]),
            origin_terminal=json.get("origin_terminal"),
            duration=json.get("duration"),
            fare_brand_name=json.get("fare_brand_name"),
            segments=get_and_transform(
                json,
                "segments",
                lambda value: [
                    OfferSliceSegment.from_json(segment) for segment in value
                ],
                [],
            ),
            conditions=OfferSliceConditions.from_json(json["conditions"]),
        )


@dataclass
class OfferPassenger:
    """The passenger travelling"""

    id: str
    age: int
    type: str
    given_name: Optional[str]
    family_name: Optional[str]
    loyalty_programme_accounts: Sequence[LoyaltyProgrammeAccount]

    allowed_types = ["adult"]

    class InvalidType(Exception):
        """Invalid passenger type provided"""

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            age=json["age"],
            type=json["type"],
            given_name=json.get("given_name"),
            family_name=json.get("family_name"),
            loyalty_programme_accounts=get_and_transform(
                json,
                "loyalty_programme_accounts",
                lambda value: [
                    LoyaltyProgrammeAccount.from_json(loyalty_programme_account)
                    for loyalty_programme_account in value
                ],
                [],
            ),
        )


@dataclass
class Offer:
    """After you've searched for flights by creating an offer request, we'll send your
    search to a range of airlines, which may return offers.

    Each offer represents flights you can buy from an airline at a particular price that
    meet your search criteria.

    You'll see slices inside the offers. Each slice will also include a list of one or
    more specific flights (called segments) that the airline is offering to get the
    passengers where they want to go.
    """

    id: str
    live_mode: bool
    allowed_passenger_identity_document_types: Sequence[str]
    available_services: Sequence[Service]
    base_amount: str
    base_currency: str
    conditions: OfferConditions
    created_at: datetime
    updated_at: datetime
    expires_at: datetime
    owner: Airline
    passenger_identity_documents_required: bool
    passengers: Sequence[OfferPassenger]
    payment_requirements: PaymentRequirements
    slices: Sequence[OfferSlice]
    tax_amount: Optional[str]
    tax_currency: Optional[str]
    total_emissions_kg: str
    total_amount: str
    total_currency: str

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            live_mode=json["live_mode"],
            allowed_passenger_identity_document_types=json[
                "allowed_passenger_identity_document_types"
            ],
            available_services=get_and_transform(
                json,
                "available_services",
                lambda value: [Service.from_json(passenger) for passenger in value],
                [],
            ),
            base_amount=json["base_amount"],
            base_currency=json["base_currency"],
            conditions=OfferConditions.from_json(json["conditions"]),
            created_at=datetime.strptime(json["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            updated_at=datetime.strptime(json["updated_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            expires_at=parse_datetime(json["expires_at"]),
            owner=Airline.from_json(json["owner"]),
            passenger_identity_documents_required=json[
                "passenger_identity_documents_required"
            ],
            passengers=get_and_transform(
                json,
                "passengers",
                lambda value: [
                    OfferPassenger.from_json(passenger) for passenger in value
                ],
                [],
            ),
            payment_requirements=PaymentRequirements.from_json(
                json["payment_requirements"]
            ),
            slices=get_and_transform(
                json,
                "slices",
                lambda value: [OfferSlice.from_json(slice) for slice in value],
                [],
            ),
            tax_amount=json.get("tax_amount"),
            tax_currency=json.get("tax_currency"),
            total_amount=json["total_amount"],
            total_currency=json["total_currency"],
            total_emissions_kg=json["total_emissions_kg"],
        )


def parse_datetime(value: str) -> datetime:
    # There are inconsistent formats used for this field depending on the
    # endpoint
    if len(value) == 20:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ")
    else:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")



from dataclasses import dataclass
from typing import Optional, Sequence

from duffel_api.utils import get_and_transform


@dataclass
class SeatMapCabinRowSectionElementSeatService:
    """A seat for a passenger. If the available_services list is empty (which will be
    represented as an empty list : []), the seat is unavailable.

    For display, all seats should be displayed with the same static width.

    """

    id: str
    passenger_id: str
    total_amount: str
    total_currency: str

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            passenger_id=json["passenger_id"],
            total_amount=json["total_amount"],
            total_currency=json["total_currency"],
        )


@dataclass
class SeatMapCabinWings:
    """Where the wings of the aircraft are in relation to rows in the cabin.

    The numbers correspond to the indices of the first and the last row which are
    overwing. You can use this to draw a visual representation of the wings to help
    users get a better idea of what they will see outside their window.

    The indices are 0th-based and are for all rows, not just those that have seats.

    This is null when no rows of the cabin are overwing.
    """

    first_row_index: int
    last_row_index: int

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            first_row_index=json["first_row_index"],
            last_row_index=json["last_row_index"],
        )


@dataclass
class SeatMapCabinRowSectionElement:
    """The element that makes up a section"""

    type: str
    # Only for seat element
    designator: Optional[str]
    name: Optional[str]
    disclosures: Optional[Sequence[str]]
    available_services: Optional[Sequence[SeatMapCabinRowSectionElementSeatService]]

    allowed_types = [
        "seat",
        "bassinet",
        "empty",
        "exit_row",
        "lavatory",
        "galley",
        "closet",
        "stairs",
    ]

    class InvalidType(Exception):
        """Invalid seat map cabin row section element provided"""

    def __post_init__(self):
        if self.type not in SeatMapCabinRowSectionElement.allowed_types:
            raise SeatMapCabinRowSectionElement.InvalidType(self.type)

        return self

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            type=json["type"],
            designator=json.get("designator"),
            name=json.get("name"),
            disclosures=json.get("disclosures"),
            available_services=get_and_transform(
                json,
                "available_services",
                lambda value: [
                    SeatMapCabinRowSectionElementSeatService.from_json(service)
                    for service in value
                ],
            ),
        )


@dataclass
class SeatMapCabinRowSection:
    """Each row is divided into sections by one or more aisles."""

    elements: Sequence[SeatMapCabinRowSectionElement]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            elements=get_and_transform(
                json,
                "elements",
                lambda value: [
                    SeatMapCabinRowSectionElement.from_json(element)
                    for element in value
                ],
                [],
            ),
        )


@dataclass
class SeatMapCabinRow:
    """Row sections are broken up by aisles. Rows are ordered from front to back of the
    aircraft.
    """

    sections: Sequence[SeatMapCabinRowSection]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            sections=get_and_transform(
                json,
                "sections",
                lambda value: [
                    SeatMapCabinRowSection.from_json(section) for section in value
                ],
                [],
            ),
        )


@dataclass
class SeatMapCabin:
    """Cabins are ordered by deck from lowest to highest, and then within each deck from
    the front to back of the aircraft.
    """

    cabin_class: str
    deck: int
    wings: Optional[SeatMapCabinWings]
    aisles: int
    rows: Sequence[SeatMapCabinRow]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            cabin_class=json["cabin_class"],
            deck=json["deck"],
            wings=get_and_transform(json, "wings", SeatMapCabinWings.from_json),
            aisles=json["aisles"],
            rows=get_and_transform(
                json,
                "rows",
                lambda value: [SeatMapCabinRow.from_json(row) for row in value],
                [],
            ),
        )


@dataclass
class SeatMap:
    """Seat maps are used to build a rich experience for your customers so they can select
    a seat as part of an order.

    A seat map includes the data for rendering seats in the relevant cabins, along with
    their total cost and other information such as disclosures.

    A seat is a special kind of service in that they're not shown when getting an
    individual offer with return_available_services set to true. They're only available
    through this endpoint.

    So far we support selecting seats when you create an order. This means we do not
    support selecting a seat after an order has already been created or cancelling a
    booked seat.

    Display recommendations
    =======================

    Each seat, empty, and bassinet element should be displayed with the same static width,
    while other elements should ideally fill or shrink in the available space in the
    section. However, displaying them with a static width could also be appropriate.

    If these elements don't fill the whole section, they should be displayed as
    middle-aligned by default.
    """

    id: str
    slice_id: str
    segment_id: str
    cabins: Sequence[SeatMapCabin]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            slice_id=json["slice_id"],
            segment_id=json["segment_id"],
            cabins=get_and_transform(
                json,
                "cabins",
                lambda value: [SeatMapCabin.from_json(cabin) for cabin in value],
                [],
            ),
        )



from ...http_client import HttpClient, Pagination
from ...models import Offer, OfferPassenger


class OfferClient(HttpClient):
    """Client to interact with Offers"""

    def __init__(self, **kwargs):
        self._url = "/air/offers"
        super().__init__(**kwargs)

    class InvalidOfferId(Exception):
        """Invalid offer ID data provided"""

    class InvalidOfferPassengerId(Exception):
        """Invalid offer passenger ID data provided"""

    class InvalidFamilyName(Exception):
        """Invalid family name data provided"""

    class InvalidGivenName(Exception):
        """Invalid given name data provided"""

    class MissingLoyaltyProgrammeAccountValue(Exception):
        """A value was was missing in the data provided"""

    class InvalidLoyaltyProgrammeAirlineIataCode(Exception):
        """Invalid loyalty programme account airline IATA code data provided"""

    class InvalidLoyaltyProgrammeAccountNumber(Exception):
        """Invalid loyalty programme account account number data provided"""

    @staticmethod
    def _validate_update_passenger_args(
        offer_id,
        offer_passenger_id,
        family_name,
        given_name,
        loyalty_programme_accounts,
    ):
        """Validates passenger details"""

        if not offer_id.strip():
            raise OfferClient.InvalidOfferId(offer_id)
        if not offer_passenger_id.strip():
            raise OfferClient.InvalidOfferPassengerId(offer_passenger_id)
        if not family_name.strip():
            raise OfferClient.InvalidFamilyName(family_name)
        if not given_name.strip():
            raise OfferClient.InvalidGivenName(given_name)
        for loyalty_programme_account in loyalty_programme_accounts:
            if "airline_iata_code" not in loyalty_programme_account:
                raise OfferClient.MissingLoyaltyProgrammeAccountValue(
                    "airline_iata_code"
                )
            if "account_number" not in loyalty_programme_account:
                raise OfferClient.MissingLoyaltyProgrammeAccountValue("account_number")
            if not loyalty_programme_account["airline_iata_code"].strip():
                raise OfferClient.InvalidLoyaltyProgrammeAirlineIataCode(
                    loyalty_programme_account["airline_iata_code"]
                )
            if not loyalty_programme_account["account_number"].strip():
                raise OfferClient.InvalidLoyaltyProgrammeAccountNumber(
                    loyalty_programme_account["account_number"]
                )

    def get(self, id_, return_available_services=False):
        """GET /air/offers/:id"""
        params = {}
        if return_available_services:
            params["return_available_services"] = "true"

        response = self.do_get(f"{self._url}/{id_}", query_params=params)
        if response is not None:
            return Offer.from_json(response["data"])

    def list(self, offer_request_id, sort=None, max_connections=None, limit=50):
        """GET /air/offers"""
        params = {"limit": limit, "offer_request_id": offer_request_id}
        if sort:
            params["sort"] = sort
        if max_connections:
            params["max_connections"] = max_connections
        return Pagination(self, Offer, params)

    def update_passenger(
        self,
        offer_id,
        offer_passenger_id,
        family_name,
        given_name,
        loyalty_programme_accounts,
    ):
        """PATCH /air/offers/:offer_id/passengers/:passenger_id"""

        OfferClient._validate_update_passenger_args(
            offer_id,
            offer_passenger_id,
            family_name,
            given_name,
            loyalty_programme_accounts,
        )

        url = f"{self._url}/{offer_id}/passengers/{offer_passenger_id}"
        body = {
            "data": {
                "loyalty_programme_accounts": [
                    {
                        "airline_iata_code": loyalty_programme_account[
                            "airline_iata_code"
                        ],
                        "account_number": loyalty_programme_account["account_number"],
                    }
                    for loyalty_programme_account in loyalty_programme_accounts
                ],
                "given_name": given_name,
                "family_name": family_name,
            }
        }

        res = self.do_patch(url, body=body)
        if res is not None:
            return OfferPassenger.from_json(res["data"])



from ...http_client import HttpClient
from ...models import OrderChange


class OrderChangeClient(HttpClient):
    """Entrypoint to create and confirm order changes."""

    class InvalidPayment(Exception):
        """Invalid payment data provided"""

    class InvalidPaymentType(Exception):
        """Invalid payment type provided"""

    @staticmethod
    def _validate_payment(payment):
        """Validate payment for the order change."""
        if set(payment.keys()) != set(["amount", "currency", "type"]):
            raise OrderChangeClient.InvalidPayment(payment)
        if payment["type"] not in ["arc_bsp_cash", "balance", "payments"]:
            raise OrderChangeClient.InvalidPaymentType(payment["type"])

    def __init__(self, **kwargs):
        """Instantiate an order change client."""
        self._url = "/air/order_changes"
        super().__init__(**kwargs)

    def get(self, id_):
        """GET /air/order_changes/:id."""
        res = self.do_get(f"{self._url}/{id_}")
        if res is not None:
            return OrderChange.from_json(res["data"])

    def create(self, selected_order_change_offer):
        """Create a pending order change.

        To begin the process of changing an order you need to create an order
        change. The OrderChange will contain the `selected_order_change_offer`
        reference of the change you wish to make to your order.

        To proceed, you must confirm the change using the Confirm an order
        change endpoint.
        """
        res = self.do_post(
            self._url,
            body={"data": {"selected_order_change_offer": selected_order_change_offer}},
        )
        if res is not None:
            return OrderChange.from_json(res["data"])

    def confirm(self, id_, payment_):
        """Confirm an order change.

        Once you've created a pending order change, you'll know the
        `change_total_amount` due for the change.

        To actually change the order, you'll need to confirm the change. The
        booking with the airline will be updated with the new slice you
        previously chose, and the change_total_amount will be charged to your
        specific payment type.

        If the amount of change_total_amount is negative, then this will be
        returned to the refund_to method (e.g. your Duffel balance). You'll
        then need to refund your customer (e.g. back to their credit/debit
        card).
        """
        OrderChangeClient._validate_payment(payment_)

        url = f"{self._url}/{id_}/actions/confirm"

        res = self.do_post(
            url,
            body={
                "data": {
                    "payment": {
                        "amount": payment_["amount"],
                        "currency": payment_["currency"],
                        "type": payment_["type"],
                    }
                }
            },
        )
        if res is not None:
            return OrderChange.from_json(res["data"])



from duffel_api.models import SeatMap

from .fixtures import raw_fixture


def test_seat_map_model_parsing():
    name = "get-seat-maps"
    with raw_fixture(name) as fixture:
        assert SeatMap.from_json(fixture["data"][0])



import datetime

from .fixtures import fixture


def test_get_order_change_offer_by_id(requests_mock):
    url = "air/order_change_offers/id"
    with fixture("get-order-change-offer-by-id", url, requests_mock.get, 200) as client:
        order_change_offer = client.order_change_offers.get("id")

        assert order_change_offer.id == "oco_0000A3vUda8dKRtUSQPSXw"
        assert order_change_offer.change_total_amount == "90.80"
        assert order_change_offer.change_total_currency == "GBP"
        assert order_change_offer.created_at == datetime.datetime(
            2020, 1, 17, 10, 12, 14, 545000
        )
        assert order_change_offer.expires_at == datetime.datetime(
            2020, 1, 17, 10, 42, 14
        )
        assert order_change_offer.new_total_amount == "35.50"
        assert order_change_offer.new_total_currency == "GBP"
        assert order_change_offer.order_change_id == "oce_0000A4QasEUIjJ6jHKfhHU"
        assert order_change_offer.penalty_amount == "10.50"
        assert order_change_offer.penalty_currency == "GBP"
        assert order_change_offer.refund_to == "arc_bsp_cash"
        assert order_change_offer.updated_at == datetime.datetime(
            2020, 1, 17, 10, 12, 14, 545000
        )
        assert isinstance(order_change_offer.slices.add, type([]))
        assert isinstance(order_change_offer.slices.remove, type([]))
        order_change_offer.slices.add[0].id == "sli_00009htYpSCXrwaB9Dn123"
        order_change_offer.slices.remove[0].id == "sli_00009htYpSCXrwaB9Dn123"


def test_get_order_change_offers(requests_mock):
    # We need a way to ensure pagination finished in a mocking environment
    end_pagination_url = (
        "http://someaddress/air/order_change_offers?limit=50"
        + "&order_change_request_id=ocr_0000A3bQP9RLVfNUcdpLpw"
        + "&after=g2wAAAACbQAAABBBZXJvbWlzdC1LaGFya2l2bQAAAB%3D"
    )
    end_pagination_response = {"meta": {"after": None}, "data": []}
    requests_mock.get(
        end_pagination_url, complete_qs=True, json=end_pagination_response
    )

    url = (
        "air/order_change_offers?limit=50"
        + "&order_change_request_id=ocr_0000A3bQP9RLVfNUcdpLpw"
    )
    with fixture(
        "get-order-change-offers-by-order-change-request-id",
        url,
        requests_mock.get,
        200,
    ) as client:
        paginated_order_change_offers = client.order_change_offers.list(
            "ocr_0000A3bQP9RLVfNUcdpLpw"
        )
        order_change_offers = list(paginated_order_change_offers)
        assert len(order_change_offers) == 1
        offer = order_change_offers[0]
        assert offer.id == "oco_0000A3vUda8dKRtUSQPSXw"
        assert offer.refund_to == "original_form_of_payment"



from dataclasses import dataclass
from typing import Optional


@dataclass
class Airline:
    """Airlines are used to identify the air travel companies selling and operating
    flights
    """

    id: str
    name: str
    iata_code: Optional[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            name=json["name"],
            iata_code=json.get("iata_code"),
        )



from dataclasses import dataclass
from datetime import date, datetime
from typing import Optional, Sequence, Union

from duffel_api.models import Airport, City, LoyaltyProgrammeAccount, Offer
from duffel_api.utils import get_and_transform


@dataclass
class OfferRequestSlice:
    """One-way journeys can be expressed using one slice, whereas return trips will need
    two.
    """

    destination_type: str
    destination: Union[Airport, City]
    origin_type: str
    origin: Union[Airport, City]
    departure_date: date

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""

        destination_type = json["destination_type"]
        destination = json["destination"]
        if destination_type == "airport":
            destination = Airport.from_json(destination)
        elif destination_type == "city":
            destination = City.from_json(destination)

        origin_type = json["origin_type"]
        origin = json["origin"]
        if origin_type == "airport":
            origin = Airport.from_json(origin)
        elif origin_type == "city":
            origin = City.from_json(origin)

        return cls(
            destination_type=destination_type,
            destination=destination,
            origin_type=origin_type,
            origin=origin,
            departure_date=date.fromisoformat(json["departure_date"]),
        )


@dataclass
class OfferRequestPassenger:
    """The passengers who want to travel"""

    id: str
    age: int
    type: str
    given_name: Optional[str]
    family_name: Optional[str]
    loyalty_programme_accounts: Sequence[LoyaltyProgrammeAccount]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            age=json["age"],
            type=json["type"],
            given_name=json.get("given_name"),
            family_name=json.get("family_name"),
            loyalty_programme_accounts=get_and_transform(
                json,
                "loyalty_programme_accounts",
                lambda value: [
                    LoyaltyProgrammeAccount.from_json(loyalty_programme_account)
                    for loyalty_programme_account in value
                ],
                [],
            ),
        )


@dataclass
class OfferRequest:
    """To search for flights, you'll need to create an offer request. An offer request
    describes the passengers and where and when they want to travel (in the form of a
    list of slices). It may also include additional filters (e.g. a particular cabin to
    travel in).
    """

    id: str
    live_mode: bool
    cabin_class: str
    created_at: datetime
    offers: Sequence[Offer]
    slices: Sequence[OfferRequestSlice]
    passengers: Sequence[OfferRequestPassenger]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            live_mode=json["live_mode"],
            cabin_class=json["cabin_class"],
            offers=get_and_transform(
                json,
                "offers",
                lambda value: [Offer.from_json(offer) for offer in value],
                [],
            ),
            slices=get_and_transform(
                json,
                "slices",
                lambda value: [OfferRequestSlice.from_json(slice) for slice in value],
                [],
            ),
            passengers=get_and_transform(
                json,
                "passengers",
                lambda value: [
                    OfferRequestPassenger.from_json(passenger) for passenger in value
                ],
                [],
            ),
            created_at=datetime.strptime(json["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
        )



from datetime import datetime

import pytest

from duffel_api.api import OrderCreate, OrderUpdate

from .fixtures import fixture


def test_get_order_by_id(requests_mock):
    url = "air/orders/id"
    with fixture("get-order-by-id", url, requests_mock.get, 200) as client:
        order = client.orders.get("id")
        assert order.id == "ord_00009hthhsUZ8W4LxQgkjo"
        assert not order.live_mode
        assert order.payment_status.awaiting_payment
        assert len(order.passengers) == 1
        assert order.synced_at == datetime(2020, 4, 11, 15, 48, 11)
        assert len(order.slices) == 1
        slice = order.slices[0]
        assert slice.conditions.change_before_departure.allowed
        assert slice.origin_type == "airport"
        assert len(order.slices[0].segments) == 1
        assert len(order.slices[0].segments[0].passengers) == 1
        passenger = order.slices[0].segments[0].passengers[0]
        assert len(passenger.baggages) == 1
        assert passenger.baggages[0].quantity == 1
        assert passenger.baggages[0].type == "checked"
        assert passenger.cabin_class == "economy"
        assert passenger.cabin_class_marketing_name == "Economy Basic"
        assert passenger.passenger_id == "passenger_0"
        assert passenger.seat is not None
        assert passenger.seat.designator == "14B"
        assert len(passenger.seat.disclosures) == 2
        assert passenger.seat.disclosures[0] == "Do not seat children in exit row seats"
        assert (
            passenger.seat.disclosures[1]
            == "Do not seat passengers with special needs in exit row seats"
        )
        assert passenger.seat.name == "Exit row seat"


def test_get_orders(requests_mock):
    # We need a way to ensure pagination finished in a mocking environment
    end_pagination_url = (
        "http://someaddress/air/orders?limit=50"
        + "&after=g2wAAAACbQAAABBBZXJvbWlzdC1LaGFya2l2bQAAAB%3D"
    )
    end_pagination_response = {"meta": {"after": None}, "data": []}
    requests_mock.get(
        end_pagination_url, complete_qs=True, json=end_pagination_response
    )

    url = "air/orders?limit=50"
    with fixture("get-orders", url, requests_mock.get, 200) as client:
        paginated_orders = client.orders.list()
        orders = list(paginated_orders)
        assert len(orders) == 1
        order = orders[0]
        assert order.id == "ord_00009hthhsUZ8W4LxQgkjo"
        assert len(order.slices) == 1
        assert len(order.slices[0].segments) == 1
        assert len(order.slices[0].segments[0].passengers) == 1
        passenger = order.slices[0].segments[0].passengers[0]
        assert len(passenger.baggages) == 1
        assert passenger.baggages[0].quantity == 1
        assert passenger.baggages[0].type == "checked"
        assert passenger.cabin_class == "economy"
        assert passenger.cabin_class_marketing_name == "Economy Basic"
        assert passenger.passenger_id == "passenger_0"
        assert passenger.seat is None
        assert order.synced_at == datetime(2020, 4, 11, 15, 48, 11)


def test_create_instant_order(requests_mock):
    url = "air/orders"
    with fixture("create-instant-order", url, requests_mock.post, 201) as client:
        passengers = [
            {
                "born_on": "2000-02-21",
                "email": "3124@example.com",
                "family_name": "Doe",
                "given_name": "Jane",
                "gender": "f",
                "id": "pas_00009hj8USM7Ncg31cBCLL",
                "phone_number": "00333333333",
                "title": "miss",
            }
        ]
        payments = [{"amount": "32", "currency": "GBP", "type": "balance"}]
        creation = client.orders.create()
        order = (
            creation.passengers(passengers)
            .selected_offers(["offer-id"])
            .payments(payments)
            .execute()
        )
        assert order.id == "ord_00009hthhsUZ8W4LxQgkjo"
        assert order.payment_status.awaiting_payment is False
        assert order.payment_status.payment_required_by is None
        assert order.payment_status.price_guarantee_expires_at is None
        assert order.synced_at == datetime(2020, 4, 11, 15, 48, 11)
        assert len(order.services) == 1
        service = order.services[0]
        assert service.id == "ser_00009UhD4ongolulWd9123"


def test_create_hold_order(requests_mock):
    url = "air/orders"
    with fixture("create-hold-order", url, requests_mock.post, 201) as client:
        passengers = [
            {
                "born_on": "2000-01-12",
                "email": "8567@example.com",
                "family_name": "Doe",
                "given_name": "Jack",
                "gender": "m",
                "id": "pas_00009hj8USM7Ncg31cBANZ",
                "phone_number": "00333333333",
                "title": "mr",
            }
        ]
        creation = client.orders.create()
        order = (
            creation.passengers(passengers)
            .selected_offers(["offer-id"])
            .hold()
            .execute()
        )
        assert order.id == "ord_00009hthhsUZ8W4LxQgkjo"
        assert order.payment_status.awaiting_payment is True
        assert order.payment_status.payment_required_by == datetime(
            2020, 1, 17, 10, 42, 14
        )
        assert order.payment_status.price_guarantee_expires_at == datetime(
            2020, 1, 17, 10, 42, 14
        )
        assert order.synced_at == datetime(2020, 4, 11, 15, 48, 11)
        assert len(order.services) == 1
        service = order.services[0]
        assert service.id == "ser_00009UhD4ongolulWd9123"


def test_create_order_with_invalid_data(requests_mock):
    url = "air/orders"
    with fixture("create-instant-order", url, requests_mock.post, 422) as client:
        creation = client.orders.create()
        with pytest.raises(OrderCreate.InvalidNumberOfPassengers):
            creation.execute()

        with pytest.raises(OrderCreate.InvalidPassenger):
            creation.passengers([{}]).execute()

        passengers = [
            {
                "born_on": "2000-02-21",
                "email": "3124@example.com",
                "family_name": "Doe",
                "given_name": "Jane",
                "gender": "f",
                "id": "pas_00009hj8USM7Ncg31cBCLL",
                "phone_number": "00333333333",
                "title": "miss",
            }
        ]

        selected_offers = ["offer-id"]
        payments = [{"amount": "32", "currency": "GBP", "type": "balance"}]

        # from this point we have valid passengers
        creation = creation.passengers(passengers)
        with pytest.raises(OrderCreate.InvalidNumberOfPayments):
            creation.execute()

        with pytest.raises(OrderCreate.InvalidPayment):
            creation.payments([{}]).execute()

        # from this point we have valid payments
        creation = creation.payments(payments)
        with pytest.raises(OrderCreate.InvalidSelectedOffersLength):
            creation.execute()

        with pytest.raises(OrderCreate.InvalidService):
            creation.selected_offers(selected_offers).services([{}]).execute()


def test_update_order(requests_mock):
    url = "air/orders/ord_00009hthhsUZ8W4LxQgkjo"
    with fixture("update-order-by-id", url, requests_mock.patch, 200) as client:
        order = (
            client.orders.update("ord_00009hthhsUZ8W4LxQgkjo")
            .metadata(
                {
                    "customer_prefs": "window seat",
                    "payment_intent_id": "pit_00009htYpSCXrwaB9DnUm2",
                }
            )
            .execute()
        )

        assert order.id == "ord_00009hthhsUZ8W4LxQgkjo"
        assert order.synced_at == datetime(2020, 4, 11, 15, 48, 11)
        assert order.metadata == {
            "customer_prefs": "window seat",
            "payment_intent_id": "pit_00009htYpSCXrwaB9DnUm2",
        }


def test_update_order_with_invalid_data(requests_mock):
    url = "air/orders/ord_00009hthhsUZ8W4LxQgkjo"
    with fixture("update-order-by-id", url, requests_mock.patch, 422) as client:
        updating = client.orders.update("ord_00009hthhsUZ8W4LxQgkjo")

        with pytest.raises(OrderUpdate.InvalidMetadata):
            updating.metadata(1).execute()



from dataclasses import dataclass
from datetime import datetime
from typing import Optional


@dataclass
class Payment:
    """To pay for an unpaid order you've previously created, you'll need to create a
    payment for it.

    The payment details you provide will be charged and, if successful, your order will
    be confirmed with the airline.

    """

    id: str
    live_mode: bool
    type: str
    amount: str
    currency: Optional[str]
    created_at: datetime

    allowed_types = ["arc_bsp_cash", "balance"]

    class InvalidType(Exception):
        """Invalid payment type provided"""

    def __post_init__(self):
        if self.type not in Payment.allowed_types:
            raise Payment.InvalidType(self.type)

        return self

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            type=json["type"],
            live_mode=json["live_mode"],
            amount=json["amount"],
            currency=json.get("currency"),
            created_at=datetime.strptime(json["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
        )



from duffel_api.models import Offer

from .fixtures import raw_fixture


def test_offer_model_parsing():
    name = "get-offer-by-id"
    with raw_fixture(name) as fixture:
        assert Offer.from_json(fixture["data"])


def test_offer_model_parsing_with_nullable_payment_requirements():
    name = "get-offer-by-id-with-null-payment-requirements"
    with raw_fixture(name) as fixture:
        assert Offer.from_json(fixture["data"])



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="duffel-api",
    version="0.4.1",
    author="Duffel Engineering",
    author_email="client-libraries@duffel.com",
    description="Client library for the Duffel API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/duffelhq/duffel-api-python",
    packages=find_packages(),
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
    ],
    keywords="duffel api flights airports airlines aircraft",
    python_requires=">=3.7",
    install_requires=["requests>=2.25"],
)



"""Http Client, api response and error management"""
import os

from requests import Request, Session
from requests import codes as http_codes

from .utils import version


class ClientError(Exception):
    """An error originated from the client"""


class ApiError(Exception):
    """An error originated from the API"""

    def __init__(self, headers, json):
        self._headers = headers
        self.meta = json["meta"]
        self.errors = json["errors"]
        # We always only print the first error message
        self.message = self.errors[0]["message"]
        super().__init__(self.message)

    def __str__(self):
        """Custom message for our API errors"""
        err = self.errors[0]
        return f"{err['type']}: {err['title']}: {self.message}"

    @property
    def headers(self):
        """Request headers"""
        return self._headers


# TODO(nlopes): I don't like this. Pagination in this way means the user will be
# constrained by this flow.  A better way would be to return a ListObject that contains
# the first list of objects and then if the user wants auto pagination, they can call
# something like `auto_paginate` or something along those lines.  We also don't provide
# good mechanisms for the user to react to rate limiting when auto paginating.
class Pagination:
    """A way to do pagination on list() calls"""

    def __init__(self, client, caller, params):
        self._client = client
        self._caller = caller

        if params["limit"] > 200:
            # We're vaguely faking the structure of the error structure returned
            # from the API.
            raise ApiError([], {"errors": [{"message": "limit exceeds 200"}]})
        self._params = params

    def __iter__(self):
        """Iterate over the response items and yield one by one"""
        response = self._client.do_get(
            self._client._url,
            query_params=self._params,
        )

        while "meta" in response:
            after = response["meta"]["after"]
            for entry in response["data"]:
                yield self._caller.from_json(entry)

            if after is None:
                break

            self._params["after"] = after
            response = self._client.do_get(
                self._client._url,
                query_params=self._params,
            )


class HttpClient:
    """Http Client to manage all calls to the Duffel API"""

    URL = "https://api.duffel.com"
    VERSION = "v1"

    def __init__(self, access_token=None, api_url=None, api_version=None, **settings):
        if api_url is not None:
            self._api_url = api_url
        else:
            self._api_url = HttpClient.URL
        if api_version is not None:
            self._api_version = api_version
        else:
            self._api_version = HttpClient.VERSION

        self.http_session = Session()
        self._settings = settings

        user_agent = f"Duffel/{self._api_version} duffel_api_python/{version()}"
        self.http_session.headers.update({"User-Agent": user_agent})
        self.http_session.headers.update({"Accept": "application/json"})
        self.http_session.headers.update({"Duffel-Version": self._api_version})
        if not access_token:
            access_token = os.getenv("DUFFEL_ACCESS_TOKEN")
            if not access_token:
                raise ClientError("must set DUFFEL_ACCESS_TOKEN")
        self.http_session.headers.update({"Authorization": f"Bearer {access_token}"})

    def _http_call(self, endpoint, method, query_params=None, body=None):
        """Perform the http call and wrap the response in a ApiError in case an error
        occurred

        """
        request_url = self._api_url + endpoint
        request = Request(method, request_url, params=query_params, json=body)
        prepared = self.http_session.prepare_request(request)
        response = self.http_session.send(prepared, **self._settings)
        if response.status_code in [
            http_codes.ok,
            http_codes.created,
        ]:
            try:
                return response.json()
            except ValueError as err:
                raise Exception(f"something bad happened: {response.text}") from err
        elif response.status_code == http_codes.no_content:
            return None
        else:
            try:
                raise ApiError(response.headers, response.json())
            except ValueError as err:
                raise Exception(f"something bad happened: {response.text}") from err
            raise response.raise_for_status()

    def do_get(self, endpoint, method="GET", query_params=None, body=None):
        """Issue a GET request to `endpoint`"""
        return self._http_call(endpoint, method, query_params, body)

    def do_post(self, endpoint, method="POST", query_params=None, body=None):
        """Issue a POST request to `endpoint`"""
        return self._http_call(endpoint, method, query_params, body)

    def do_delete(self, endpoint, method="DELETE", query_params=None, body=None):
        """Issue a DELETE request to `endpoint`"""
        return self._http_call(endpoint, method, query_params, body)

    def do_put(self, endpoint, method="PUT", query_params=None, body=None):
        """Issue a PUT request to `endpoint`"""
        return self._http_call(endpoint, method, query_params, body)

    def do_patch(self, endpoint, method="PATCH", query_params=None, body=None):
        """Issue a PATCH request to `endpoint`"""
        return self._http_call(endpoint, method, query_params, body)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Sequence

from duffel_api.models import Aircraft, Airline, Place
from duffel_api.utils import get_and_transform


@dataclass
class OrderChangeSlicesSliceSegment:
    """A segment within a slice that is being removed or added"""

    id: str
    aircraft: Optional[Aircraft]
    arriving_at: datetime
    departing_at: datetime
    destination: Place
    destination_terminal: Optional[str]
    origin: Place
    origin_terminal: Optional[str]
    distance: Optional[str]
    duration: Optional[str]
    marketing_carrier: Airline
    marketing_carrier_flight_number: str
    operating_carrier: Airline
    operating_carrier_flight_number: Optional[str]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            aircraft=get_and_transform(json, "aircraft", Aircraft.from_json),
            arriving_at=datetime.strptime(json["arriving_at"], "%Y-%m-%dT%H:%M:%S"),
            departing_at=datetime.strptime(json["departing_at"], "%Y-%m-%dT%H:%M:%S"),
            destination=Place.from_json(json["destination"]),
            destination_terminal=json.get("destination_terminal"),
            origin=Place.from_json(json["origin"]),
            origin_terminal=json.get("origin_terminal"),
            distance=json.get("distance"),
            duration=json.get("duration"),
            marketing_carrier=Airline.from_json(json["marketing_carrier"]),
            marketing_carrier_flight_number=json["marketing_carrier_flight_number"],
            operating_carrier=Airline.from_json(json["operating_carrier"]),
            operating_carrier_flight_number=json.get("operating_carrier_flight_number"),
        )


@dataclass
class OrderChangeSlicesSlice:
    """A slice that is being removed or added"""

    id: str
    duration: Optional[str]
    destination: Place
    destination_type: str
    origin: Place
    origin_type: str
    segments: Sequence[OrderChangeSlicesSliceSegment]

    allowed_place_types = ["airport", "city"]

    class InvalidPlaceType(Exception):
        """Invalid type of place"""

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            destination_type=json["destination_type"],
            destination=Place.from_json(json["destination"]),
            origin_type=json["origin_type"],
            origin=Place.from_json(json["origin"]),
            duration=json.get("duration"),
            segments=get_and_transform(
                json,
                "segments",
                lambda value: [
                    OrderChangeSlicesSliceSegment.from_json(segment)
                    for segment in value
                ],
                [],
            ),
        )


@dataclass
class OrderChangeSlices:
    """The slices to be added and/or removed"""

    add: Sequence[OrderChangeSlicesSlice]
    remove: Sequence[OrderChangeSlicesSlice]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            add=get_and_transform(
                json,
                "add",
                lambda value: [
                    OrderChangeSlicesSlice.from_json(slice) for slice in value
                ],
                [],
            ),
            remove=get_and_transform(
                json,
                "remove",
                lambda value: [
                    OrderChangeSlicesSlice.from_json(slice) for slice in value
                ],
                [],
            ),
        )


@dataclass
class OrderChange:
    """Once you've created an order change request, and you've chosen which
    slices to add and remove, you'll then want to create an order change.

    To create an order change, you just need to provide the order change offer
    ID.

    If change_total_amount is greater than zero, you will need to provide
    payment details when confirming the change.

    If change_total_amount is less than zero, this amount will be returned to
    the payment method specified by refund_to.
    """

    id: str
    live_mode: bool
    order_id: str
    expires_at: datetime
    change_total_amount: str
    change_total_currency: str
    new_total_amount: str
    new_total_currency: str
    penalty_total_amount: str
    penalty_total_currency: str
    refund_to: str
    slices: OrderChangeSlices
    created_at: datetime
    confirmed_at: Optional[datetime]

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            id=json["id"],
            order_id=json["order_id"],
            live_mode=json["live_mode"],
            expires_at=parse_datetime(json["expires_at"]),
            confirmed_at=get_and_transform(json, "confirmed_at", parse_datetime),
            change_total_amount=json["change_total_amount"],
            change_total_currency=json["change_total_currency"],
            new_total_amount=json["new_total_amount"],
            new_total_currency=json["new_total_currency"],
            penalty_total_amount=json["penalty_total_amount"],
            penalty_total_currency=json["penalty_total_currency"],
            refund_to=json["refund_to"],
            created_at=datetime.strptime(json["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            slices=OrderChangeSlices.from_json(json["slices"]),
        )


def parse_datetime(value: str) -> datetime:
    # There are inconsistent formats used for this field depending on the
    # endpoint
    if len(value) == 20:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ")
    else:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")



from ...http_client import HttpClient, Pagination
from ...models import Airline


class AirlineClient(HttpClient):
    """Client to interact with Airlines"""

    def __init__(self, **kwargs):
        self._url = "/air/airlines"
        super().__init__(**kwargs)

    def get(self, id_):
        """GET /air/airlines/:id"""
        res = self.do_get(f"{self._url}/{id_}")
        if res is not None:
            return Airline.from_json(res["data"])

    def list(self, limit=50):
        """GET /air/airlines"""
        return Pagination(self, Airline, {"limit": limit})



"""Assorted auxiliary functions"""
from typing import Any


def identity(value: Any) -> Any:
    """Given a value, return the exact same value"""
    return value


def get_and_transform(dict: dict, key: str, fn, default=None):
    """Get a value from a dictionary and transform it or return
    None if the key isn't present"""
    try:
        value = dict[key]
        if value is None:
            return value
        else:
            return fn(value)
    except KeyError:
        return default


def version() -> str:
    """Return the version specified in the package (setup.py) during runtime"""
    import pkg_resources

    return pkg_resources.require("duffel_api")[0].version



from dataclasses import dataclass


@dataclass
class LoyaltyProgrammeAccount:
    """A passenger's loyalty programme account"""

    airline_iata_code: str
    account_number: str

    @classmethod
    def from_json(cls, json: dict):
        """Construct a class instance from a JSON response."""
        return cls(
            airline_iata_code=json["airline_iata_code"],
            account_number=json["account_number"],
        )
